import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
import math


class phc_gc_hex(pya.PCellDeclarationHelper):
    """
    Input: length, width
    """

    import numpy

    def __init__(self):
        # Important: initialize the super class
        super(phc_gc_hex, self).__init__()

        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.243)
        self.param("x", self.TypeInt, "Number of holes in x direction", default=78)
        self.param("y", self.TypeInt, "Number of holes in y direction", default=50)
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.0735)
        self.param("vertices", self.TypeInt, "Number of vertices in circle", default=32)
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("positive", self.TypeInt, "Positive", default=False)
        self.param("apodized", self.TypeInt, "apodized", default=False)
        self.param(
            "feature_size",
            self.TypeDouble,
            "minimum feature size (microns)",
            default=0.06,
        )
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])
        self.param(
            "invert", self.TypeLayer, "Layer to invert", default=TECHNOLOGY["Si"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "phc_gc_hex_a%s-r%.3f" % (self.a, self.r)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        debug = False
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)
        LayerInvert = ly.layer(self.invert)
        n_vertices = int(self.vertices)
        a = self.a / dbu
        r = self.r / dbu
        n_x = int(math.ceil(self.x / 2))
        n_y = int(math.ceil(self.y / 2))
        positive = bool(self.positive)
        minimum_feature = self.feature_size
        apodized = bool(self.apodized)
        length_slab_x = 2 * n_x * a

        length_slab_y = 2 * (n_y - 2) * a

        k = 2

        # function to creat polygon pts for right half of a hole in a hexagon unit cell

        def circle(x, y, r):
            npts = n_vertices
            theta = 2 * math.pi / npts  # increment, in radians
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (y + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        def hexagon_hole_half(a, r):
            npts = 10
            theta_div = math.pi / 3
            theta_div_hole = math.pi / npts
            triangle_length = a / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            for i in range(0, npts + 1):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            r * math.cos(math.pi / 2 - i * theta_div_hole),
                            r * math.sin(math.pi / 2 - i * theta_div_hole),
                        )
                    )
                )
            return pts

        def hexagon_shifthole_half(a, r):
            npts = 10
            theta_div = math.pi / 3
            theta_div_hole = math.pi / npts
            triangle_length = a * 1.235 / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            for i in range(0, npts + 1):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            r * math.cos(math.pi / 2 - i * theta_div_hole),
                            r * math.sin(math.pi / 2 - i * theta_div_hole),
                        )
                    )
                )
            return pts

        # function to creat polygon pts for right half of a hexagon unit cell
        def hexagon_half(a):
            theta_div = math.pi / 3
            triangle_length = a / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            return pts

        # create the right and left half of the hole and hexagon cells
        # hole_cell = pya.Region()
        # hexagon_cell = pya.Region()

        # Define Si slab and hole region for future subtraction
        Si_slab = pya.Region()
        Si_slab.insert(
            pya.Box(
                -length_slab_x / 2 + a * 2,
                -length_slab_y / 2,
                length_slab_x / 2 - a,
                length_slab_y / 2,
            )
        )
        hole = pya.Region()
        hole_r = r

        # function to generate points to create a circle
        def circle(x, y, r):
            npts = n_vertices
            theta = 2 * math.pi / npts  # increment, in radians
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (y + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        import numpy as np

        def gaussian(x, mu, sig):
            return (
                1.0
                / (np.sqrt(2.0 * np.pi) * sig)
                * np.exp(-np.power((x - mu) / sig, 2.0) / 2)
            )

        # raster through all holes with shifts and waveguide

        for j in range(-n_y, n_y + 1):
            if j % 2 == 0:
                skip = 0
                apodization = 0
                for i in range(-n_x, n_x + 1):
                    if i == 0:
                        continue
                    if skip == 0:
                        skip = 1
                        continue
                    elif skip == 1:
                        skip = 2
                        continue
                    elif skip == 2:
                        skip = 3
                        apodization = apodization + 1
                        continue
                    elif skip == 3:
                        skip = 1

                    # radius=r/gaussian(float(i)/(2*(n_x+1)),2*(n_x+1),10)
                    location = float(i)
                    location = abs(location)
                    # radius=r/gaussian(location/(2*(n_x+1)),2*(n_x+1),0.02)
                    # radius=(((2*n_x)-abs(i+n_x+3)*r)/(2*n_x))
                    radius = (float(apodization) / ((n_x * 2 / 3) - 1)) * r
                    if radius < minimum_feature * 500:
                        radius = minimum_feature * 500
                    if apodized == False:
                        radius = r

                        # radius=minimum_feature
                    hole_cell = circle(0, 0, radius)
                    hole_poly = pya.Polygon(hole_cell)
                    if debug:
                        print("x1 " + str(apodization))
                    hole_x = abs(i) / i * (abs(i) - 0.5) * a
                    hole_y = j * a * math.sqrt(3) / 2
                    hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                    hole_t = hole_poly.transformed(hole_trans)
                    hole.insert(hole_t)
            elif j % 2 == 1:
                skipodd = 0
                apodization = 0
                for i in range(-n_x, n_x + 1):
                    if i == -n_x:
                        continue
                    if i == n_x:
                        continue
                    if skipodd == 0:
                        skipodd = 1
                        continue
                    elif skipodd == 1:
                        skipodd = 2
                        apodization = apodization + 1
                        continue
                    elif skipodd == 2:
                        skipodd = 3
                    elif skipodd == 3:
                        skipodd = 1

                    # radius=(((2*n_x)-abs(i+n_x+3)*r)/(2*n_x))
                    radius = (float(apodization) / ((n_x * 2 / 3) - 1)) * r
                    if radius < minimum_feature * 500:
                        radius = minimum_feature * 500
                    if apodized == False:
                        radius = r

                        # radius=minimum_feature
                    # radius=r*(float(abs(i))/(2*(n_x+1)))
                    hole_cell = circle(0, 0, radius)
                    hole_poly = pya.Polygon(hole_cell)
                    if debug:
                        print("x2 " + str(apodization))
                        print("p " + str(n_x))
                    hole_x = i * a
                    hole_y = j * a * math.sqrt(3) / 2
                    hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                    hole_t = hole_poly.transformed(hole_trans)
                    hole.insert(hole_t)
        if positive == True:
            phc = hole
        else:
            phc = Si_slab - hole
        self.cell.shapes(LayerSiN).insert(phc)

        # print(hole_t_0)
        box_l = a / 2

        # Pins on the waveguide:
        pin_length = 200
        pin_w = a

        t = pya.Trans(Trans.R0, -length_slab_x / 2 + a * 2, 0)
        pin = pya.Path(
            [pya.Point(-pin_length / 2, 0), pya.Point(pin_length / 2, 0)], pin_w
        )
        pin_t = pin.transformed(t)
        self.cell.shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text("pin1", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        points = [[-length_slab_x / 2 + a * 2, 0], [length_slab_x / 2 - a, 0]]
        points = [Point(each[0], each[1]) for each in points]
        path = Path(points, length_slab_y)
        self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())

        if positive == True:
            self.cell.shapes(LayerInvert).insert(path.simple_polygon())

        return

        hole = pya.Region()

        hole_cell_pts = hexagon_hole_half(a, r)
        hexagon_pts = hexagon_half(a)
        hole_shiftcell_pts = hexagon_shifthole_half(a, r)
        hole_cell_poly_0 = pya.Polygon(hole_cell_pts)
        hexagon_cell_poly_0 = pya.Polygon(hexagon_pts)
        hole_shiftcell_poly_0 = pya.Polygon(hole_shiftcell_pts)

        hole_trans = pya.Trans(pya.Trans.R180)
        hole_cell_poly_1 = hole_cell_poly_0.transformed(hole_trans)
        hexagon_cell_poly_1 = hexagon_cell_poly_0.transformed(hole_trans)
        hole_shiftcell_poly_1 = hole_shiftcell_poly_0.transformed(hole_trans)
        skip = 1
        skip2 = 0
        # create the photonic crystal with shifts and waveguides
        for j in range(-y + 1, y):
            if j % 2 == 0:
                for i in range(-x, x + 1):
                    # waveguide
                    if j == k and i > 3:
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
                    # filling the edges with half cell

                    elif i != 0:
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
            elif j % 2 == 1:
                for i in range(-x, x + 1):
                    if skip == 0 and i % 3 == 0:
                        skip = 1
                        continue
                    if skip == 1:
                        skip = 2
                        continue
                    if skip == 2:
                        skip = 0

                    if i == -x:
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_1)
                    if i == x:
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
                        hole.insert(hole_t_0)
                    # waveguide
                    if j == k and i > 3:
                        hole_x = i * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
                    # filling the edges with half cell
                    elif i in (-x, x):
                        hole_x = i * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        if i == -x:
                            hole_t = hole_cell_poly_0.transformed(hole_trans)
                        else:
                            hole_t = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t)

                    else:
                        hole_x = i * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)

        # print(hole_t_0)
        box_l = a / 2

        cover_box = pya.Box(-length_slab_x / 2, -a / 2, length_slab_x / 2, a / 2)
        box_y = y * a * math.sqrt(3) / 2
        cover_box_trans_0 = pya.Trans(Trans.R0, 0, box_y)
        cover_box_trans_1 = pya.Trans(Trans.R0, 0, -box_y)
        cover_box_t_0 = cover_box.transformed(cover_box_trans_0)
        cover_box_t_1 = cover_box.transformed(cover_box_trans_1)
        # hole.insert(pya.Box())
        self.cell.shapes(LayerSiN).insert(hole)
        self.cell.shapes(LayerSiN).insert(cover_box_t_0)
        self.cell.shapes(LayerSiN).insert(cover_box_t_1)
        # Pins on the waveguide:
        pin_length = 200
        pin_w = a

        t = pya.Trans(Trans.R0, -length_slab_x / 2, 0)
        pin = pya.Path(
            [pya.Point(-pin_length / 2, 0), pya.Point(pin_length / 2, 0)], pin_w
        )
        pin_t = pin.transformed(t)
        self.cell.shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text("pin1", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        points = [[-length_slab_x / 2, 0], [length_slab_x / 2, 0]]
        points = [Point(each[0], each[1]) for each in points]
        path = Path(points, length_slab_y)
        self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
import math
from SiEPIC._globals import PIN_LENGTH as pin_length
from SiEPIC.extend import to_itype
from SiEPIC.scripts import path_to_waveguide, connect_pins_with_waveguide, connect_cell
from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech, get_layout_variables

class ebeam_dream_microlens_edge_couplers_BB(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the ebeam_dream_lens_edge_couplers_BB
    
    Authors: Dream Photonics
    """

    def __init__(self):

        # Important: initialize the super class
        super(ebeam_dream_microlens_edge_couplers_BB, self).__init__()

        self.technology_name = 'EBeam'
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # declare the parameters
        self.param("num_channels", self.TypeInt, "Number of Channels (1 - 16)", default = 1)
        p = self.param("center_wavelength", self.TypeList, "Center Wavelength of Operation [nm]", default = '1550')
        p.add_choice('1550 nm', '1550')
        p.add_choice('1310 nm', '1310')

        self.param("ref_wg", self.TypeBoolean, "Include reference waveguide", default=True)

        #declare the layers
        self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'], hidden=True)
        self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
        self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
        self.param("fibertarget", self.TypeLayer, "Fiber Target Layer", default=TECHNOLOGY['FbrTgt'])
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY['Text'])
        self.param("bb",self.TypeLayer,"BB Layer", default=TECHNOLOGY['BlackBox'])

    def can_create_from_shape_impl(self):
        return False


    def produce(self, layout, layers, parameters, cell):
        # This is the main part of the implementation: create the layout
        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        #fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes
        
        LayerSiN = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerFbrTgtN = ly.layer(self.fibertarget)
        LayerTEXTN = ly.layer(self.textl)
        LayerBBN = ly.layer(self.bb)


        w_tip = to_itype(0.100,dbu)
        l_taper = to_itype(65,dbu)
        num_channels = self.num_channels
        offset = to_itype(0,dbu)
        pitch = to_itype(127,dbu)
        Lw2 = to_itype(15,dbu)
        Lw3 = Lw2 + to_itype(20,dbu)
        wavelength = int(self.center_wavelength)
        waveguide_type = 'Strip TE 1550 nm, w=500 nm'

        if num_channels < 1:
            num_channels = 1
        if num_channels > 16:
            num_channels = 16

        if wavelength == 1310:
            waveguide_type = 'Strip TE 1310 nm, w=350 nm'
            w_waveguide = to_itype(0.350,dbu)
        elif wavelength == 1550:
            waveguide_type = 'Strip TE 1550 nm, w=500 nm'
            w_waveguide = to_itype(0.500,dbu)


        def circle(x,y,r):
            npts = 180
            theta = 2*math.pi/npts
            pts = []
            for i in range(0,npts):
              pts.append(pya.Point.from_dpoint(pya.DPoint((x+r*math.cos(i*theta))/1,(y+r*math.sin(i*theta))/1)))
            return pts

        #draw one loopback device
        if self.ref_wg:
            for ref_loop in range(2):
                #draw fibre target circle
                align_circle = circle(offset,-pitch*(ref_loop+1),2/dbu)
                #place fibre target circle
                shapes(LayerFbrTgtN).insert(pya.Polygon(align_circle))

        if self.ref_wg:
            #create waveguide to for loopback
            loopback_path = pya.DPath([pya.DPoint(0,-127),pya.DPoint((offset + l_taper + Lw2)*dbu+15,-127),pya.DPoint((offset + l_taper + Lw2)*dbu+15,-254),pya.DPoint(0,-254)],0.5)
            self.layout.technology_name = self.technology_name #required otherwise "create_cell" doesn't load
            pcell = self.layout.create_cell("Waveguide",self.technology_name,{"path": loopback_path, "waveguide_type": waveguide_type})
            t = pya.Trans(pya.Trans.R0,0,0)
            self.cell.copy(pcell,LayerSiN,LayerBBN)
            wg = self.cell.insert(pya.CellInstArray(pcell.cell_index(),t))
            wg.flatten()
            self.cell.clear(LayerDevRecN)
            self.cell.clear(LayerPinRecN)
            self.cell.clear(LayerSiN)


        ##########################################################################################################################################################################
        #draw N tapers
        for n_ch in range(int(num_channels)):

            #draw the taper
            taper_pts = [pya.Point(0,-w_waveguide/2+pitch*n_ch),pya.Point(0,w_waveguide/2+pitch*n_ch),pya.Point(offset + l_taper + Lw3,w_waveguide/2+pitch*n_ch),pya.Point(offset + l_taper + Lw3,-w_waveguide/2+pitch*n_ch)]

            #place the taper
            shapes(LayerBBN).insert(pya.Polygon(taper_pts))

            #draw and place pin on the waveguide:
            x = offset + l_taper + Lw3
            t = pya.Trans(pya.Trans.R0, x, pitch*n_ch)
            pin = pya.Path([pya.Point(-pin_length/2,0),pya.Point(pin_length/2,0)], w_waveguide)
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text(f"opt{n_ch+1}",t)
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 3/dbu
   
            #draw fibre target circle
            align_circle = circle(offset,pitch*n_ch,2/dbu)

            #place fibre target circle
            shapes(LayerFbrTgtN).insert(pya.Polygon(align_circle))

        #draw devrec box
        if self.ref_wg:
            devrec_pts = [pya.Point(0,pitch*n_ch+30/dbu),pya.Point(x,pitch*n_ch+30/dbu),pya.Point(x,-pitch*2-30/dbu),pya.Point(0,-pitch*2-30/dbu)]
        else:
            devrec_pts = [pya.Point(0,pitch*n_ch+30/dbu),pya.Point(x,pitch*n_ch+30/dbu),pya.Point(x,-30/dbu),pya.Point(0,-30/dbu)]
        
        #place devrec box
        shapes(LayerDevRecN).insert(pya.Polygon(devrec_pts))

        #edge of chip text
        t = pya.Trans(pya.Trans.R0,0,1/dbu)
        text = pya.Text("<- Edge of chip",t)
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 3/dbu

        #BB description
        t = pya.Trans(pya.Trans.R0,0,-15/dbu)
        text = pya.Text("  Number of Channel(s): " + str(num_channels) + "\n  Center Wavelength: " + str(wavelength) + " nm",t)
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 3/dbu

        #BB description
        t = pya.Trans(pya.Trans.R0, 0,-25/dbu)
        text = pya.Text("<- 25 MFD lens",t)
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 3/dbu

        #draw DP BB logo
        import os
        dir_path = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../gds/EBeam_Dream/"))
        filename = os.path.join(dir_path, 'DP_lens_BB_logo.gds')

        ly2=pya.Layout()
        ly2.read(filename)
        top_cell = ly2.top_cell()
        top_cell.flatten(True)
        self.cell.copy_shapes(top_cell)

        #draw lenses
        width_lens = to_itype(50, dbu)
        length_lens = to_itype(50, dbu)

        if self.ref_wg:
            for n_ch in range(int(num_channels+2)):
                lens_pts = [pya.Point(0,-width_lens/2+pitch*n_ch-2*pitch), pya.Point(0,width_lens/2+pitch*n_ch-2*pitch), pya.Point(-length_lens,width_lens/2+pitch*n_ch-2*pitch),pya.Point(-length_lens,-width_lens/2+pitch*n_ch-2*pitch)]
                shapes(LayerBBN).insert(pya.Polygon(lens_pts))
                lens = circle(-length_lens,pitch*n_ch-2*pitch,25/dbu)
                shapes(LayerBBN).insert(pya.Polygon(lens))
        else:
            for n_ch in range(int(num_channels)):
                lens_pts = [pya.Point(0,-width_lens/2+pitch*n_ch), pya.Point(0,width_lens/2+pitch*n_ch), pya.Point(-length_lens,width_lens/2+pitch*n_ch),pya.Point(-length_lens,-width_lens/2+pitch*n_ch)]
                shapes(LayerBBN).insert(pya.Polygon(lens_pts))
                lens = circle(-length_lens,pitch*n_ch,25/dbu)
                shapes(LayerBBN).insert(pya.Polygon(lens))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_dream_microlens_edge_couplers_%s_%s_BB" % (
            self.center_wavelength,
            self.num_channels,
        )




import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class BraggWaveguide_holes(pya.PCellDeclarationHelper):
    """

    Waveguide Bragg grating using holes

    """

    def __init__(self):
        # Important: initialize the super class
        super(BraggWaveguide_holes, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param(
            "number_of_periods", self.TypeInt, "Number of grating periods", default=10
        )
        self.param(
            "grating_period", self.TypeDouble, "Grating period (microns)", default=0.260
        )
        self.param("fill_factor", self.TypeDouble, "Grating fill factor", default=0.5)
        self.param(
            "hole_width", self.TypeDouble, "Corrugration width (microns)", default=0.07
        )
        self.param("wg_width", self.TypeDouble, "Waveguide width", default=0.35)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "BraggWaveguide_holes_%s-%.3f-%.3f-%.3f" % (
            self.number_of_periods,
            self.grating_period,
            self.hole_width,
            self.fill_factor,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype

        # Draw the Bragg grating:
        grating_period = to_itype(self.grating_period, dbu)
        w = to_itype(self.wg_width, dbu)
        hole_width = to_itype(self.hole_width, dbu)
        fill_factor = self.fill_factor

        for i in range(0, self.number_of_periods):
            x = i * grating_period
            box1 = Box(x, -w / 2, x + grating_period * (1 - fill_factor), w / 2)
            box2 = Box(
                x + grating_period * (1 - fill_factor),
                -hole_width / 2,
                x + grating_period,
                -w / 2,
            )
            box3 = Box(
                x + grating_period * (1 - fill_factor),
                hole_width / 2,
                x + grating_period,
                w / 2,
            )
            shapes(LayerSiN).insert(box1)
            shapes(LayerSiN).insert(box2)
            shapes(LayerSiN).insert(box3)
            length = x + grating_period

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        t = Trans(Trans.R0, 0, 0)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("opt1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.3 / dbu

        t = Trans(Trans.R0, length, 0)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("opt2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.3 / dbu
        shape.text_halign = 2

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 10, 0)
        text = Text("Component=ebeam_bragg_te1550", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 9, 0)
        text = Text(
            "Spice_param:number_of_periods=%s grating_period=%.3g hole_width=%.3g fill_factor=%.3g"
            % (
                self.number_of_periods,
                self.grating_period * 1e-6,
                self.hole_width * 1e-6,
                self.fill_factor,
            ),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        t = Trans(Trans.R0, 0, 0)
        path = Path([Point(0, 0), Point(length, 0)], 3 * w)
        shapes(LayerDevRecN).insert(path.simple_polygon())

        # print('Done: BraggWaveguide_holes')



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Contra_DC_dualBragg(pya.PCellDeclarationHelper):
    """
    Author:   Mustafa Hammood
              Mustafa@siepic.com
    """

    def __init__(self):
        # Important: initialize the super class
        super(Contra_DC_dualBragg, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param(
            "number_of_periods",
            self.TypeInt,
            "Number of grating periods (sidewall)",
            default=300,
        )
        self.param(
            "number_of_periods_swg",
            self.TypeInt,
            "Number of grating periods (sub-wavelength)",
            default=260,
        )
        self.param(
            "grating_period1",
            self.TypeDouble,
            "Grating period (microns)",
            default=0.318,
        )
        self.param(
            "grating_period2",
            self.TypeDouble,
            "Subwavelength Grating period (microns)",
            default=0.420,
        )
        self.param("gap", self.TypeDouble, "Gap (microns)", default=0.15)
        self.param(
            "corrugation_width1",
            self.TypeDouble,
            "Waveguide 1 Corrugration width (microns)",
            default=0.03,
        )
        self.param(
            "corrugation_width2",
            self.TypeDouble,
            "Waveguide 2 Corrugration width (microns)",
            default=0.04,
        )
        self.param(
            "corrugation_width",
            self.TypeDouble,
            "Subwavelength Corrugration width (microns)",
            default=0.04,
        )
        self.param("AR", self.TypeBoolean, "Anti-Reflection Coating", default=True)
        self.param(
            "sinusoidal",
            self.TypeBoolean,
            "Grating Type (Rectangular=False, Sinusoidal=True)",
            default=False,
        )
        self.param("wg1_width", self.TypeDouble, "Waveguide 1 width", default=0.45)
        self.param("wg2_width", self.TypeDouble, "Waveguide 2 width", default=0.55)
        self.param("index", self.TypeDouble, "Gaussian Index", default=2.8)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Contra_DC_dualBragg%s-%.3f" % (
            self.number_of_periods,
            self.grating_period1,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        import math

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype

        # Draw the Bragg grating (bottom):
        box_width = to_itype(self.grating_period1 / 2, dbu)
        SWG_width = to_itype(self.grating_period2 / 2, dbu)

        grating_period1 = to_itype(self.grating_period1, dbu)
        grating_period2 = to_itype(self.grating_period2, dbu)

        w = to_itype(self.wg1_width, dbu)
        GaussianIndex = self.index
        half_w = w / 2
        half_corrugation_w = to_itype(self.corrugation_width1 / 2, dbu)

        N = self.number_of_periods

        vertical_offset1 = to_itype(self.gap / 2 + self.wg1_width / 2, dbu)
        vertical_offset2 = to_itype(-self.gap / 2 - self.wg2_width / 2, dbu)

        for i in range(0, self.number_of_periods):
            x = int(round((i * self.grating_period1) / dbu))

            profileFunction = math.exp(
                -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
            )
            profile = int(round(self.corrugation_width1 / 2 / dbu)) * profileFunction

            box1 = Box(
                x,
                vertical_offset1 - half_w - profile,
                x + box_width,
                vertical_offset1 + half_w - profile,
            )

            box2 = Box(
                x + box_width,
                vertical_offset1 - half_w + profile,
                x + 2 * box_width,
                vertical_offset1 + half_w + profile,
            )

            shapes(LayerSiN).insert(box1)
            shapes(LayerSiN).insert(box2)

            length = x + grating_period1

        # Draw the Bragg grating (bottom):
        box_width = int(round(self.grating_period1 / 2 / dbu))
        grating_period = int(round(self.grating_period1 / dbu))
        w = to_itype(self.wg2_width, dbu)
        GaussianIndex = self.index
        half_w = w / 2
        half_corrugation_w = int(round(self.corrugation_width2 / 2 / dbu))

        for i in range(0, self.number_of_periods):
            x = int(round((i * self.grating_period1) / dbu))

            profileFunction = math.exp(
                -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
            )
            profile = int(round(self.corrugation_width2 / 2 / dbu)) * profileFunction

            box1 = Box(
                x,
                vertical_offset2 - half_w - profile,
                x + box_width,
                vertical_offset2 + half_w - profile,
            )

            box2 = Box(
                x + box_width,
                vertical_offset2 - half_w + profile,
                x + 2 * box_width,
                vertical_offset2 + half_w + profile,
            )

            shapes(LayerSiN).insert(box1)
            shapes(LayerSiN).insert(box2)

            length = x + grating_period1

        # Draw the Bragg grating (SWG):
        vertical_offset = (
            int(round(self.corrugation_width2 / dbu))
            - int(round(self.corrugation_width1 / dbu))
        ) / 4

        box_width = int(round(self.grating_period2 / 2 / dbu))
        grating_period = int(round(self.grating_period2 / dbu))
        w = to_itype(self.corrugation_width, dbu)
        GaussianIndex = self.index
        half_corrugation_w = int(round(self.corrugation_width / dbu))

        for i in range(0, self.number_of_periods_swg):
            x_swg = int(round((i * self.grating_period2) / dbu))

            profileFunction = math.exp(
                -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
            )
            profile = int(round(self.corrugation_width / dbu)) * profileFunction

            box1 = Box(
                x_swg,
                vertical_offset + w / 2 + profile,
                x_swg + box_width,
                vertical_offset - w / 2 - profile,
            )

            shapes(LayerSiN).insert(box1)

            length_swg = x + grating_period2

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        w = to_itype(self.wg2_width, dbu)
        t = Trans(Trans.R0, 0, vertical_offset2)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg1_width, dbu)
        t = Trans(Trans.R0, 0, vertical_offset1)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin3", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg1_width, dbu)
        t = Trans(Trans.R0, length, vertical_offset1)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg2_width, dbu)
        t = Trans(Trans.R0, length, vertical_offset2)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin4", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 10, 0)
        text = Text("Component=ebeam_contra_dc", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 9, 0)
        text = Text(
            "Spice_param:number_of_periods=%s grating_period=%.3fu corrugation_width=%.3fu sinusoidal=%s"
            % (
                self.number_of_periods,
                self.grating_period1,
                self.corrugation_width1,
                int(self.sinusoidal),
            ),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        t = Trans(Trans.R0, 0, 0)
        path = Path([Point(0, 0), Point(length, 0)], 3 * w)
        shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class SWG_bentWG(pya.PCellDeclarationHelper):
    def __init__(self):
        # Important: initialize the super class
        super(SWG_bentWG, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("pitch", self.TypeDouble, "Period [um]", default=0.2)
        self.param("w", self.TypeDouble, "Waveguide Width [um]", default=0.5)
        self.param(
            "r", self.TypeDouble, "Radius [um]", default=5
        )  # minimum radius is 5 um in TE
        self.param("ff", self.TypeDouble, "Duty Cycle", default=0.5)
        self.param("angle", self.TypeDouble, "Angle", default=360)
        self.param("gap", self.TypeDouble, "Bus Gap [um]", default=0.06)
        self.param("W_ratio", self.TypeDouble, "Lin/Lout", default=1)
        self.param(
            "doublebus",
            self.TypeBoolean,
            "Bus Type (Single=False, Double=True)",
            default=False,
        )
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "SWG_bentWG_%.3f" % (self.pitch)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        debug = False

        from SiEPIC.extend import to_itype
        import math
        from pya import DPolygon
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes
        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # variables
        pitch = self.pitch
        w = self.w
        r = self.r
        ff = self.ff
        angle = self.angle
        gap = self.gap
        W_ratio = self.W_ratio
        doublebus = self.doublebus
        if debug:
            print(doublebus)

        # set the deltaL of sT-shaped silicon pillar (deltaL = (Lin-Lout)/4)
        deltaL = (2 * pitch * ff) * (W_ratio - 1) / ((1 + W_ratio) * (4))

        if r - w / 2 <= 0:
            r = 5
            if debug:
                print("invalid radius, set r to default: 5")

        # Calculate number of segments
        s1 = pitch * ff  # silicon
        s2 = pitch - s1  # gap

        # calculate best radius
        pi = math.pi
        const = math.ceil(2 * pi * r / (s1 + s2))
        # if doesn't divide evenly, replace r with best possible r
        if (2 * pi * r) % (s1 + s2) != 0:
            r = const * (s1 + s2) / (2 * pi)
            if debug:
                print("r adjusted to " + str(r) + "um to fit periods perfectly.")

        # theta1 = (s1/r)
        # theta2 = (s2/r)
        theta1 = math.atan(s1 / r)  # angle of silicon compared to the origin point
        theta2 = math.atan(s2 / r)  # angle of the gap compared to the origin point
        nSeg = int(
            math.floor(angle / (math.degrees(theta1) + math.degrees(theta2)))
        )  # how many segments to have
        si_first = True  # for alternating between silicon and gap
        j = 0  # index of how many silicon thetas
        jj = 0  # index of how many gap thetas
        ORDER = True  # ordering of the coordinates for polygon drawing

        # xo = [(r-w/2)*math.cos(0)]
        # yo = [(r-w/2)*math.sin(0)]
        # xo.append((r+w/2)*math.cos(0))
        # yo.append((r+w/2)*math.sin(0))

        xo = [(r - w / 2) * math.cos(0) + deltaL * math.sin(0)]
        yo = [(r - w / 2) * math.sin(0) - deltaL * math.cos(0)]
        xo.append((r + w / 2) * math.cos(0) - deltaL * math.sin(0))
        yo.append((r + w / 2) * math.sin(0) + deltaL * math.cos(0))

        for i in range(0, nSeg * 2):
            if si_first:
                j = j + 1
                si_first = not (si_first)
            else:
                jj = jj + 1
                si_first = not (si_first)

            if ORDER:
                xo.append(
                    (r + w / 2) * math.cos(j * theta1 + jj * theta2)
                    + deltaL * math.sin(j * theta1 + jj * theta2)
                )
                yo.append(
                    (r + w / 2) * math.sin(j * theta1 + jj * theta2)
                    - deltaL * math.cos(j * theta1 + jj * theta2)
                )
                xo.append(
                    (r - w / 2) * math.cos(j * theta1 + jj * theta2)
                    - deltaL * math.sin(j * theta1 + jj * theta2)
                )
                yo.append(
                    (r - w / 2) * math.sin(j * theta1 + jj * theta2)
                    + deltaL * math.cos(j * theta1 + jj * theta2)
                )
                ORDER = not (ORDER)
            else:
                xo.append(
                    (r - w / 2) * math.cos(j * theta1 + jj * theta2)
                    + deltaL * math.sin(j * theta1 + jj * theta2)
                )
                yo.append(
                    (r - w / 2) * math.sin(j * theta1 + jj * theta2)
                    - deltaL * math.cos(j * theta1 + jj * theta2)
                )
                xo.append(
                    (r + w / 2) * math.cos(j * theta1 + jj * theta2)
                    - deltaL * math.sin(j * theta1 + jj * theta2)
                )
                yo.append(
                    (r + w / 2) * math.sin(j * theta1 + jj * theta2)
                    + deltaL * math.cos(j * theta1 + jj * theta2)
                )
                ORDER = not (ORDER)

            if len(xo) == 4:
                dpts = [pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)
                xo = []
                yo = []

        # Draw Bus WG
        # calulate ideal length of bus
        bus_length = self.cell.bbox().height() * dbu + pitch * 2
        constant = math.ceil(bus_length / (s1 + s2))
        if bus_length % (s1 + s2) != 0:
            bus_length = constant * (s1 + s2)

        # draw first box at center
        xo = [
            (r + w / 2 + gap),
            (r + w / 2 + gap + w),
            (r + w / 2 + gap + w),
            (r + w / 2 + gap),
        ]
        yo = [0, 0, s1, s1]
        dpts = [pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
        shapes(LayerSiN).insert(element)

        for i in range(0, int(math.ceil((constant) / 2))):
            yu = [yo[j] + i * pitch for j in range(len(yo))]
            yd = [yo[j] - i * pitch for j in range(len(yo))]

            dpts = [pya.DPoint(xo[i], yu[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)
            dpts = [pya.DPoint(xo[i], yd[i]) for i in range(len(xo))]
            dpolygon = DPolygon(dpts)
            element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
            shapes(LayerSiN).insert(element)

        if doublebus:
            for i in range(0, int(math.ceil((constant) / 2))):
                x2 = [xo[j] * -1 for j in range(len(xo))]
                yu = [yo[j] + i * pitch for j in range(len(yo))]
                yd = [yo[j] - i * pitch for j in range(len(yo))]

                dpts = [pya.DPoint(x2[i], yu[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)
                dpts = [pya.DPoint(x2[i], yd[i]) for i in range(len(xo))]
                dpolygon = DPolygon(dpts)
                element = Polygon.from_dpoly(dpolygon * (1.0 / dbu))
                shapes(LayerSiN).insert(element)

        # DEV BOX
        if doublebus is False:
            half_l = (self.cell.bbox().width() - w / dbu - gap / dbu) / 2
            half_r = self.cell.bbox().width() - half_l
            dev = Box(-half_l, yd[-4] / dbu, half_r, yu[-1] / dbu)
            shapes(LayerDevRecN).insert(dev)
            dev_width = self.cell.bbox().width() / 2
        else:
            half_l = (self.cell.bbox().width()) / 2
            half_r = self.cell.bbox().width() - half_l
            dev = Box(-half_l, yd[-4] / dbu, half_r, yu[-1] / dbu)
            shapes(LayerDevRecN).insert(dev)
            dev_width = self.cell.bbox().width() / 2

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        w = to_itype(self.w, dbu)
        gap = to_itype(self.gap, dbu)
        bus_length = to_itype(bus_length / 2, dbu)

        # Pin1
        t = Trans(Trans.R0, half_r - w / 2, yu[-1] / dbu)
        pin = Path([Point(0, -pin_length / 2), Point(0, pin_length / 2)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin2
        t = Trans(Trans.R0, half_r - w / 2, yd[-4] / dbu)
        pin = Path([Point(0, pin_length / 2), Point(0, -pin_length / 2)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        if doublebus is True:
            half_l = self.cell.bbox().width() / 2
            half_r = self.cell.bbox().width() - half_l
            # pin 3
            t = Trans(Trans.R0, -half_l + w / 2, yu[-1] / dbu)
            pin = Path([Point(0, -pin_length / 2), Point(0, pin_length / 2)], w)
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = Text("pin3", t)
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.4 / dbu

            # Pin4
            t = Trans(Trans.R0, -half_l + w / 2, yd[-4] / dbu)
            pin = Path([Point(0, pin_length / 2), Point(0, -pin_length / 2)], w)
            pin_t = pin.transformed(t)
            shapes(LayerPinRecN).insert(pin_t)
            text = Text("pin4", t)
            shape = shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.4 / dbu



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
import math


class phc_H0c_new(pya.PCellDeclarationHelper):
    """
    Input: length, width
    """

    import numpy

    def __init__(self):
        # Important: initialize the super class
        super(phc_H0c_new, self).__init__()

        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.744)
        self.param(
            "n", self.TypeInt, "Number of holes in x and y direction", default=30
        )
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.179)
        self.param(
            "wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default=3
        )
        self.param("S1x", self.TypeDouble, "S1x shift", default=0.28)
        self.param("S2x", self.TypeDouble, "S2x shift", default=0.193)
        self.param("S3x", self.TypeDouble, "S3x shift", default=0.194)
        self.param("S4x", self.TypeDouble, "S4x shift", default=0.162)
        self.param("S5x", self.TypeDouble, "S5x shift", default=0.113)
        self.param("S1y", self.TypeDouble, "S1y shift", default=-0.016)
        self.param("S2y", self.TypeDouble, "S2y shift", default=0.134)
        self.param(
            "bus_number", self.TypeInt, "2 for double, 1 for single, max 2", default=2
        )
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "H0 Cavity_a%s-r%.3f-wg_dis%.3f-n%.3f" % (
            self.a,
            self.r,
            self.wg_dis,
            self.n,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        bus_n = self.bus_number

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        a = self.a / dbu
        r = self.r / dbu
        wg_dis = self.wg_dis + 1
        n = int(math.ceil(self.n / 2))
        Sx = [self.S1x, self.S2x, self.S3x, self.S4x, self.S5x]
        Sy = [self.S1y, 0, self.S2y]
        if wg_dis % 2 == 0:
            length_slab_x = 2 * n * a
        else:
            length_slab_x = (2 * n + 1) * a

        length_slab_y = 2 * (n - 2) * a

        if bus_n == 2:
            k = -1
        else:
            k = 1

        # function to creat polygon pts for right half of a hole in a hexagon unit cell
        def hexagon_hole_half(a, r):
            npts = 10
            theta_div = math.pi / 3
            theta_div_hole = math.pi / npts
            triangle_length = a / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            for i in range(0, npts + 1):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            r * math.cos(math.pi / 2 - i * theta_div_hole),
                            r * math.sin(math.pi / 2 - i * theta_div_hole),
                        )
                    )
                )
            return pts

        def hexagon_shifthole_half(a, r):
            npts = 10
            theta_div = math.pi / 3
            theta_div_hole = math.pi / npts
            triangle_length = a * 1.235 / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            for i in range(0, npts + 1):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            r * math.cos(math.pi / 2 - i * theta_div_hole),
                            r * math.sin(math.pi / 2 - i * theta_div_hole),
                        )
                    )
                )
            return pts

        # function to creat polygon pts for right half of a hexagon unit cell
        def hexagon_half(a):
            theta_div = math.pi / 3
            triangle_length = a / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            return pts

        # create the right and left half of the hole and hexagon cells
        # hole_cell = pya.Region()
        # hexagon_cell = pya.Region()
        hole = pya.Region()

        hole_cell_pts = hexagon_hole_half(a, r)
        hexagon_pts = hexagon_half(a)
        hole_shiftcell_pts = hexagon_shifthole_half(a, r)
        hole_cell_poly_0 = pya.Polygon(hole_cell_pts)
        hexagon_cell_poly_0 = pya.Polygon(hexagon_pts)
        hole_shiftcell_poly_0 = pya.Polygon(hole_shiftcell_pts)

        hole_trans = pya.Trans(pya.Trans.R180)
        hole_cell_poly_1 = hole_cell_poly_0.transformed(hole_trans)
        hexagon_cell_poly_1 = hexagon_cell_poly_0.transformed(hole_trans)
        hole_shiftcell_poly_1 = hole_shiftcell_poly_0.transformed(hole_trans)

        # create the photonic crystal with shifts and waveguides
        for j in range(-n + 1, n):
            if j % 2 == 0:
                for i in range(-n, n + 1):
                    # waveguide
                    if (j == k * wg_dis and i > 3) or (j == wg_dis and i != 0):
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
                    # filling the edges with half cell
                    elif i in (-n, n) and wg_dis % 2 == 1:
                        hole_x = abs(i) / i * (abs(i) + 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        if i == -n:
                            hole_t = hole_cell_poly_0.transformed(hole_trans)
                        else:
                            hole_t = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t)
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
                    # x shifts
                    elif j == 0 and i in (1, -1, 2, -2, 3, -3, 4, -4, 5, -5):
                        hole_x = abs(i) / i * (abs(i) - 0.5 + Sx[abs(i) - 1]) * a
                        hole_y = 0
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_shiftcell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_shiftcell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
                    elif i != 0:
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
            elif j % 2 == 1:
                for i in range(-n, n + 1):
                    # waveguide
                    if (j == k * wg_dis and i > 3) or j == wg_dis:
                        hole_x = i * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hexagon_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hexagon_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
                    # filling the edges with half cell
                    elif wg_dis % 2 == 0 and i in (-n, n):
                        hole_x = i * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        if i == -n:
                            hole_t = hole_cell_poly_0.transformed(hole_trans)
                        else:
                            hole_t = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t)
                    # y shifts
                    elif i == 0 and j in (1, -1, 3, -3):
                        hole_x = 0
                        hole_y = (
                            j * a * (math.sqrt(3) / 2) + abs(j) / j * a * Sy[abs(j) - 1]
                        )
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_shiftcell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_shiftcell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)
                    else:
                        hole_x = i * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t_0 = hole_cell_poly_0.transformed(hole_trans)
                        hole_t_1 = hole_cell_poly_1.transformed(hole_trans)
                        hole.insert(hole_t_0)
                        hole.insert(hole_t_1)

        # print(hole_t_0)
        box_l = a / 2
        hole.insert(pya.Box(-box_l, -box_l, box_l, box_l))
        cover_box = pya.Box(-length_slab_x / 2, -a / 2, length_slab_x / 2, a / 2)
        box_y = n * a * math.sqrt(3) / 2
        cover_box_trans_0 = pya.Trans(Trans.R0, 0, box_y)
        cover_box_trans_1 = pya.Trans(Trans.R0, 0, -box_y)
        cover_box_t_0 = cover_box.transformed(cover_box_trans_0)
        cover_box_t_1 = cover_box.transformed(cover_box_trans_1)
        # hole.insert(pya.Box())
        self.cell.shapes(LayerSiN).insert(hole)
        self.cell.shapes(LayerSiN).insert(cover_box_t_0)
        self.cell.shapes(LayerSiN).insert(cover_box_t_1)
        # Pins on the waveguide:
        pin_length = 200
        pin_w = a
        wg_pos = a * math.sqrt(3) / 2 * wg_dis

        t = pya.Trans(Trans.R0, -length_slab_x / 2, wg_pos)
        pin = pya.Path(
            [pya.Point(-pin_length / 2, 0), pya.Point(pin_length / 2, 0)], pin_w
        )
        pin_t = pin.transformed(t)
        self.cell.shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text("pin1", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = pya.Trans(Trans.R0, length_slab_x / 2, wg_pos)
        pin_t = pin.transformed(t)
        self.cell.shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text("pin2", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # pin for drop waveguide
        t = pya.Trans(Trans.R0, length_slab_x / 2, -wg_pos)
        pin_t = pin.transformed(t)
        self.cell.shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text("pin3", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        points = [[-length_slab_x / 2, 0], [length_slab_x / 2, 0]]
        points = [Point(each[0], each[1]) for each in points]
        path = Path(points, length_slab_y)
        self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())



"""
This file is part of the SiEPIC-EBeam-PDK

Version history:

Lukas Chrostowski 2022/04/01
 - PCell takes an input of the waveguide type from a dropdown list, rather than
   detailed parameters.  Loaded from Waveguides.xml
"""

from pya import *
import pya


class Waveguide(pya.PCellDeclarationHelper):
    def __init__(self):
        # Important: initialize the super class
        super(Waveguide, self).__init__()

        from SiEPIC.utils import load_Waveguides_by_Tech

        """
    # https://github.com/KLayout/klayout/issues/879
    tech = self.layout.library().technology
    if not tech:
       tech = 'EBeam'
    self.technology_name = tech
    """
        self.technology_name = "EBeam"

        # Load all strip waveguides
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)

        # declare the parameters

        p = self.param(
            "waveguide_type",
            self.TypeList,
            "Waveguide Type",
            default=self.waveguide_types[0]["name"],
        )
        for wa in self.waveguide_types:
            p.add_choice(wa["name"], wa["name"])
        self.param(
            "path",
            self.TypeShape,
            "Path",
            default=DPath([DPoint(0, 0), DPoint(10, 0), DPoint(10, 10)], 0.5),
        )

        # self.param("length", self.TypeDouble, "Length", default = 0, readonly=True)
        """ todo - can add calculated values and parameters for user info
    self.param("radius", self.TypeDouble, "Bend Radius", default = 0, readonly=True)
    """

        self.cellName = "Waveguide"

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        if "waveguide_length" in dir(self):
            return "%s_%s_%.3f" % (
                self.cellName,
                self.waveguide_type.replace(" ", "_").replace(",", ""),
                self.waveguide_length,
            )
        else:
            return "%s_%s_%.3f" % (
                self.cellName,
                self.waveguide_type.replace(" ", "_").replace(",", ""),
                0,
            )

    def coerce_parameters_impl(self):
        #    self.length = self.waveguide_length
        pass

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def transformation_from_shape_impl(self):
        return Trans(Trans.R0, 0, 0)

    def parameters_from_shape_impl(self):
        self.path = self.shape.path

    def produce_impl(self):
        # https://github.com/KLayout/klayout/issues/879
        # tech = self.layout.library().technology

        # Make sure the technology name is associated with the layout
        #  PCells don't seem to know to whom they belong!
        if self.layout.technology_name == "":
            self.layout.technology_name = self.technology_name

        # Draw the waveguide geometry, new function in SiEPIC-Tools v0.3.90
        from SiEPIC.utils.layout import layout_waveguide4

        self.waveguide_length = layout_waveguide4(
            self.cell, self.path, self.waveguide_type, debug=False
        )

        # print("EBeam.%s: length %.3f um, complete" % (self.cellName, self.waveguide_length))


"""
import pya
from pya import *

class Waveguide(pya.PCellDeclarationHelper):

  def __init__(self):
    # Important: initialize the super class
    super(Waveguide, self).__init__()
    # declare the parameters
    from SiEPIC.utils import get_technology_by_name
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.param("path", self.TypeShape, "Path", default = DPath([DPoint(0,0), DPoint(10,0), DPoint(10,10)], 0.5))
    self.param("radius", self.TypeDouble, "Radius", default = 5)
    self.param("width", self.TypeDouble, "Width", default = 0.5)
    self.param("adiab", self.TypeBoolean, "Adiabatic", default = False)
    self.param("bezier", self.TypeDouble, "Bezier Parameter", default = 0.35)
    self.param("layers", self.TypeList, "Layers", default = ['Waveguide'])
    self.param("widths", self.TypeList, "Widths", default =  [0.5])
    self.param("offsets", self.TypeList, "Offsets", default = [0])
    self.param("CML", self.TypeString, "Compact Model Library (CML)", default = 'EBeam')
    self.param("model", self.TypeString, "CML Model name", default = 'ebeam_wg_integral_1550') 
    self.cellName="Waveguide"
    
  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "%s_%s" % (self.cellName, self.path)
  
  def coerce_parameters_impl(self):
    from SiEPIC.extend import to_itype
    print("EBeam.Waveguide coerce parameters")
    
    if 0:
        TECHNOLOGY = get_technology_by_name('EBeam')
        dbu = self.layout.dbu
        wg_width = to_itype(self.width,dbu)
        for lr in range(0, len(self.layers)):
          layer = self.layout.layer(TECHNOLOGY[self.layers[lr]])
          width = to_itype(self.widths[lr],dbu)
          # check to make sure that the waveguide with parameters are consistent in both places
          if self.layout.layer(TECHNOLOGY['Waveguide']) == layer:
            if width != wg_width:
              self.widths[lr] = self.width
          # check to make sure that the DevRec is bigger than the waveguide width
          if self.layout.layer(TECHNOLOGY['DevRec']) == layer:
            if width < wg_width:
              self.widths[lr] = self.width*2
              
  def can_create_from_shape_impl(self):
    return self.shape.is_path()

  def transformation_from_shape_impl(self):
    return Trans(Trans.R0,0,0)

  def parameters_from_shape_impl(self):
    self.path = self.shape.dpath
        
  def produce_impl(self):

    from SiEPIC.utils.layout import layout_waveguide2
    from SiEPIC.utils import  angle_vector
    from math import cos, sin, pi, sqrt
    import pya
    from SiEPIC.extend import to_itype
    
    # print("EBeam.Waveguide")
    
    from SiEPIC.utils import get_technology_by_name
    TECHNOLOGY = get_technology_by_name('EBeam')
    
    dbu = self.layout.dbu
    wg_width = to_itype(self.width,dbu)
    path = self.path.to_itype(dbu)

    if not (len(self.layers)==len(self.widths) and len(self.layers)==len(self.offsets) and len(self.offsets)==len(self.widths)):
      raise Exception("There must be an equal number of layers, widths and offsets")
    path.unique_points()
    pts = path.get_points()
    
    # Draw the waveguide geometry, new in SiEPIC-Tools v0.3.64
    waveguide_length = layout_waveguide2(TECHNOLOGY, self.layout, self.cell, self.layers, self.widths, self.offsets, pts, self.radius, self.adiab, self.bezier)

    pts = path.get_points()
    LayerPinRecN = self.layout.layer(TECHNOLOGY['PinRec'])
    
    t1 = Trans(angle_vector(pts[0]-pts[1])/90, False, pts[0])
    self.cell.shapes(LayerPinRecN).insert(Path([Point(-5, 0), Point(5, 0)], wg_width).transformed(t1))
    self.cell.shapes(LayerPinRecN).insert(Text("pin1", t1, 0.3/dbu, -1))
    
    t = Trans(angle_vector(pts[-1]-pts[-2])/90, False, pts[-1])
    self.cell.shapes(LayerPinRecN).insert(Path([Point(-5, 0), Point(5, 0)], wg_width).transformed(t))
    self.cell.shapes(LayerPinRecN).insert(Text("pin2", t, 0.3/dbu, -1))

    LayerDevRecN = self.layout.layer(TECHNOLOGY['DevRec'])

    # Compact model information
    angle_vec = angle_vector(pts[0]-pts[1])/90
    halign = 0 # left
    angle=0
    if angle_vec == 0: # horizontal
      halign = 2 # right
      angle=0
      dpt = Point(0, 0.2*wg_width)
    if angle_vec == 2: # horizontal
      halign = 0 # left
      angle = 0
      dpt=Point(0, 0.2*wg_width)
    if angle_vec == 1: # vertical
      halign = 2 # right
      angle = 1
      dpt=Point(0.2*wg_width,0)
    if angle_vec == -1: # vertical
      halign = 0 # left
      angle = 1
      dpt=Point(0.2*wg_width,0)
    pt2=pts[0] + dpt
    pt3=pts[0] - dpt
    pt4=pts[0] - 2*dpt
    pt5=pts[0] + 2*dpt
      

    t = Trans(angle, False, pt5)
    text = Text ('cellName=%s' % self.cellName, t, 0.1*wg_width, -1)
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pt3) 
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/%s' % self.CML, t, 0.1*wg_width, -1)
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pt2)
    text = Text ('Component=%s' % self.model, t, 0.1*wg_width, -1)
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pts[0])
    pts_txt = str([ [round(p.to_dtype(dbu).x,3), round(p.to_dtype(dbu).y,3)] for p in pts ]).replace(', ',',')
    text = Text ( \
      'Spice_param:wg_length=%.3fu wg_width=%.3fu points="%s" radius=%s' %\
        (waveguide_length, self.width, pts_txt,self.radius ), t, 0.1*wg_width, -1  )
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
    t = Trans(angle, False, pt4)
    text = Text ( \
      'Length=%.3fu' %(waveguide_length), t, 0.5*wg_width, -1  )
    text.halign=halign
    shape = self.cell.shapes(LayerDevRecN).insert(text)
"""



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Contra_DC_SWG(pya.PCellDeclarationHelper):
    """
    Author:   Mustafa Hammood
              Mustafa@siepic.com
    """

    def __init__(self):
        # Important: initialize the super class
        super(Contra_DC_SWG, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param(
            "number_of_periods", self.TypeInt, "Number of grating periods", default=300
        )
        self.param(
            "grating_period", self.TypeDouble, "Grating period (microns)", default=0.24
        )
        self.param("gap", self.TypeDouble, "Minimum gap (microns)", default=0.15)
        self.param(
            "corrugation_width1",
            self.TypeDouble,
            "Waveguide 1 Corrugration width (microns)",
            default=0.03,
        )
        self.param(
            "corrugation_width2",
            self.TypeDouble,
            "Waveguide 2 Corrugration width (microns)",
            default=0.04,
        )
        self.param("wg1_width", self.TypeDouble, "Waveguide 1 width", default=0.45)
        self.param("wg2_width", self.TypeDouble, "Waveguide 2 width", default=0.55)
        self.param("duty", self.TypeDouble, "Duty cycle (0-1)", default=0.5)
        self.param("a", self.TypeDouble, "Gaussian Index", default=2.7)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Contra_DC_SWG%s-%.3f" % (self.number_of_periods, self.grating_period)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        import math

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype

        N = self.number_of_periods
        grating_period = int(round(self.grating_period / dbu))
        misalignment = 0

        # Determine the period such that the waveguide length is as desired.  Slight adjustment to period
        N_boxes = N

        # Draw the Bragg grating:
        box_width = int(round(grating_period * self.duty))

        w1 = self.wg1_width / dbu
        half_w1 = w1 / 2
        w2 = self.wg2_width / dbu
        half_w2 = w2 / 2
        deltaW1_max = to_itype(self.corrugation_width1, dbu)
        deltaW2_max = to_itype(self.corrugation_width2, dbu)

        for i in range(0, N_boxes + 1):
            # apply apodization
            profileFunction = math.exp(-0.5 * (2 * self.a * (i - N / 2) / (N)) ** 2)
            deltaW1 = deltaW1_max * profileFunction
            deltaW2 = deltaW2_max * profileFunction

            vertical_offset = (
                int(round(self.wg2_width / 2 / dbu))
                + int(round(self.gap / dbu))
                + int(round(self.wg1_width / 2 / dbu))
            )  # +(-int(round(deltaW1))+int(round(deltaW2)))/2
            t = Trans(Trans.R0, 0, vertical_offset)

            if i % 2 == True:
                x = int(round((i * grating_period - box_width / 2)))
                box1_a = Box(x, -half_w1 - deltaW1, x + box_width, half_w1 - deltaW1)
                shapes(LayerSiN).insert(box1_a)

                box2_a = Box(
                    x + grating_period,
                    -half_w2 - deltaW2,
                    x + grating_period + box_width,
                    half_w2 - deltaW2,
                ).transformed(t)
                shapes(LayerSiN).insert(box2_a)

            else:
                x = int(round((i * grating_period - box_width / 2)))
                box1_b = Box(x, -half_w1, x + box_width, half_w1)
                shapes(LayerSiN).insert(box1_b)

                box2_b = Box(
                    x + grating_period,
                    -half_w2,
                    x + grating_period + box_width,
                    half_w2,
                ).transformed(t)
                shapes(LayerSiN).insert(box2_b)

        # missing periods due to misalignments
        box_final = Box(
            x + grating_period, -half_w1, x + grating_period + box_width, half_w1
        )
        shapes(LayerSiN).insert(box_final)
        box_final = Box(
            -box_width / 2, -half_w2 - deltaW2, box_width / 2, half_w2 - deltaW2
        ).transformed(t)
        shapes(LayerSiN).insert(box_final)

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        w = to_itype(self.wg1_width, dbu)
        t = Trans(Trans.R0, to_itype(-box_width / 2, dbu * 1000), -deltaW1 / 2)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg2_width, dbu)
        t = Trans(
            Trans.R0,
            to_itype(-box_width / 2, dbu * 1000),
            vertical_offset - deltaW2 / 2,
        )
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg2_width, dbu)
        t = Trans(
            Trans.R0,
            to_itype(x + grating_period + box_width, dbu * 1000),
            vertical_offset - deltaW2 / 2,
        )
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin3", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        w = to_itype(self.wg1_width, dbu)
        t = Trans(
            Trans.R0, to_itype(x + grating_period + box_width, dbu * 1000), -deltaW1 / 2
        )
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin4", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu



from pya import *
import pya


class Waveguide_Bend(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the waveguide bend.
    """

    def __init__(self):
        # Important: initialize the super class
        super(Waveguide_Bend, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("radius", self.TypeDouble, "Radius", default=10)
        self.param("wg_width", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        # hidden parameters, can be used to query this component:
        self.param(
            "p1",
            self.TypeShape,
            "DPoint location of pin1",
            default=Point(-10000, 0),
            hidden=True,
            readonly=True,
        )
        self.param(
            "p2",
            self.TypeShape,
            "DPoint location of pin2",
            default=Point(0, 10000),
            hidden=True,
            readonly=True,
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Waveguide_Bend(R=" + ("%.3f" % self.radius) + ")"

    def can_create_from_shape_impl(self):
        return False

    def produce(self, layout, layers, parameters, cell):
        """
        coerce parameters (make consistent)
        """
        self._layers = layers
        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # r: radius
        # w: waveguide width
        # length units in dbu

        from SiEPIC.utils import arc, arc_to_waveguide

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.silayer
        LayerSiN = self.silayer_layer
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        w = int(round(self.wg_width / dbu))
        r = int(round(self.radius / dbu))

        # draw the quarter-circle
        x = -r
        y = r
        # layout_arc_wg_dbu(self.cell, LayerSiN, x, y, r, w, 270, 360)
        t = Trans(Trans.R0, x, y)
        from SiEPIC import __version__

        if __version__ > "0.5.0":
            self.cell.shapes(LayerSiN).insert(
                arc_to_waveguide(arc(r, 270, 360, dbu=dbu), w).transformed(t)
            )
        else:
            self.cell.shapes(LayerSiN).insert(
                arc_to_waveguide(arc(r, 270, 360), w).transformed(t)
            )

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        # Pin on the top side:
        p2 = [Point(0, y - pin_length / 2), Point(0, y + pin_length / 2)]
        p2c = Point(0, y)
        self.set_p2 = p2c
        self.p2 = p2c
        pin = Path(p2, w)
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, 0, y)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pin on the left side:
        p1 = [Point(pin_length / 2 + x, 0), Point(-pin_length / 2 + x, 0)]
        p1c = Point(x, 0)
        self.set_p1 = p1c
        self.p1 = p1c
        pin = Path(p1, w)
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, x, 0)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        t = Trans(Trans.R0, x, y)
        if __version__ > "0.5.0":
            self.cell.shapes(LayerDevRecN).insert(
                arc_to_waveguide(arc(r, 270, 360, dbu=dbu), w * 3).transformed(t)
            )
        else:
            self.cell.shapes(LayerDevRecN).insert(
                arc_to_waveguide(arc(r, 270, 360), w * 3).transformed(t)
            )
        # layout_arc_wg_dbu(self.cell, LayerDevRecN, x, y, r, w*3, 270, 360)

        # Compact model information
        t = Trans(Trans.R0, x + r / 10, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/EBeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = r / 100
        t = Trans(Trans.R0, x + r / 10, r / 4)
        text = Text("Component=ebeam_bend_1550", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = r / 100
        t = Trans(Trans.R0, x + r / 10, r / 2)
        text = Text(
            "Spice_param:radius=%.3fu wg_width=%.3fu" % (self.radius, self.wg_width), t
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = r / 100



from . import *
import pya
from SiEPIC.utils import get_technology_by_name, arc_wg_xy
from SiEPIC._globals import PIN_LENGTH as pin_length
from pya import DPolygon


class e_skid_ring(pya.PCellDeclarationHelper):
    """
    Author: Ben Cohen (UBC), Hang Bobby Zou (UBC)
    bcohenkl@ece.ubc.ca
    Nov 9, 2023

    - This Pcell creates a single-bus ring resonator modified with e-skid concentric rings
    - Design control over number of inner/outer concentric rings. Gap between two rings and width is parameterized
    as an input list.
      - If a 0 is found in any entry of the list, the cell will automatically delete every element in the list
      - If the number of entries in the gap and width lists are different, the list with the minimum number of elements is considered

    - Controllable coupling. Bus length is designed as the outer most radius of the e-skid ring resonator
    """

    def __init__(self):
        # Important: initialize the super class
        super(e_skid_ring, self).__init__()
        self.technology_name = "EBeam"
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # Main Ring Parameters
        self.param("r", self.TypeDouble, "Ring Radii [um]", default=10)
        self.param("w", self.TypeDouble, "Ring Waveguide Width [um]", default=0.5)
        self.param("cg", self.TypeDouble, "Coupling Gap [um]", default=0.2)

        # Bus Parameters:
        self.param("bus_w", self.TypeDouble, "Bus Waveguide Width [um]", default=0.5)

        # e-skid parameter
        self.param(
            "gap_out", self.TypeList, "Outer Gap List [um] (0 if none)", default=[0]
        )
        self.param(
            "w_out", self.TypeList, "Outer Width List [um] (0 if none)", default=[0]
        )

        self.param(
            "gap_in",
            self.TypeList,
            "Inner Gap List [um] (0 if none)",
            default=[0.3, 0.3],
        )
        self.param(
            "w_in",
            self.TypeList,
            "Inner Width List [um] (0 if none)",
            default=[0.1, 0.1],
        )

        # Layer Parameters
        self.param(
            "layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"], hidden=True
        )
        self.param(
            "pinrec",
            self.TypeLayer,
            "PinRec Layer",
            default=TECHNOLOGY["PinRec"],
            hidden=True,
        )
        self.param(
            "devrec",
            self.TypeLayer,
            "DevRec Layer",
            default=TECHNOLOGY["DevRec"],
            hidden=True,
        )
        self.param(
            "oxideopen",
            self.TypeLayer,
            "Oxide Open Layer",
            default=TECHNOLOGY["Oxide open (to BOX)"],
            hidden=True,
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ring_resonator_with_e_skid_pillar"

    # Creating the layout
    def produce_impl(self):
        # Layout Parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        # Layers
        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # Main Ring Parameters
        r = self.r  # Ring radii [um]
        w = self.w  # Ring waveguide width [um]
        cg = self.cg  # Coupling gap

        # Bus Parameters:
        bus_w = self.bus_w  # Bus width [um]
        bus_l = 2 * (r + w)  # Bus length [um]

        # e-skid parameters
        gap_out = self.gap_out  # Gap between different widths for rings outside of the main ring resonator [um]
        w_out = self.w_out  # Width of each ring outside of the main ring resonator [um]
        gap_in = self.gap_in  # Gap between different widths for rings inside of the main ring resonator [um]
        w_in = self.w_in  # Width of each ring inside of the main ring resonator [um]

        # Converting the list elements from str (klayout default) to float
        # No entries in either array should be 0, deleting every entry after 0 is found
        for i in range(len(w_out)):
            w_out[i] = float(w_out[i])
            if w_out[i] == 0:
                w_out = w_out[:i]
                break

        for i in range(len(gap_out)):
            gap_out[i] = float(gap_out[i])
            if gap_out[i] == 0:
                w_out = w_out[:i]
                break

        for i in range(len(w_in)):
            w_in[i] = float(w_in[i])
            if w_in[i] == 0:
                w_in = w_in[:i]
                break

        for i in range(len(gap_in)):
            gap_in[i] = float(gap_in[i])
            if gap_in[i] == 0:
                gap_in = gap_in[:i]
                break

        # Ring Resonator
        if r != 0:
            shapes(LayerSiN).insert(arc_wg_xy(0, 0, (r + w / 2) / dbu, w / dbu, 0, 360))

        # Bus Waveguide
        bus_x = [-bus_l / 2, bus_l / 2, bus_l / 2, -bus_l / 2]
        bus_y = [r + w + cg, r + w + cg, r + w + cg + bus_w, r + w + cg + bus_w]

        dpts = [pya.DPoint(bus_x[i], bus_y[i]) for i in range(len(bus_x))]  # Core body
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon * (1 / dbu))
        shapes(LayerSiN).insert(element)

        # Adding outer concetric rings
        """
        rad0_out = r + w
        for i in range(min(len(gap_out), len(w_out))):
            ring = pya.Region()
            ring.insert(arc_wg_xy(0, 0, (rad0_out + gap_out[i] + w_out[i]/2)/dbu, w_out[i]/dbu, 0, 360))

            # CONSIDERATION FOR TAPERS
            box_subtract = 
            shapes(LayerSiN).insert(ring)

            #shapes(LayerSiN).insert(arc_wg_xy(0, 0, (rad0_out + gap_out[i] + w_out[i]/2)/dbu, w_out[i]/dbu, 0, 360))
            rad0_out = rad0_out + gap_out[i] + w_out[i]
        """

        # Adding inner concetric rings
        if r != 0:
            rad0_in = r
            for i in range(min(len(gap_in), len(w_in))):
                ring = pya.Region()
                ring.insert(
                    arc_wg_xy(
                        0,
                        0,
                        (rad0_in - gap_in[i] - w_in[i] / 2) / dbu,
                        w_in[i] / dbu,
                        0,
                        360,
                    )
                )
                shapes(LayerSiN).insert(ring)

                rad0_in = rad0_in - gap_in[i] - w_in[i]

        # Pins for mating:
        t = pya.DTrans(
            pya.Trans.R90, min(bus_x) / dbu, (min(bus_y) + max(bus_y)) / 2 / dbu
        )
        pin = pya.Path(
            [pya.DPoint(0, -pin_length / 2), pya.DPoint(0, pin_length / 2)], bus_w / dbu
        )
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text(
            "opt1",
            pya.DTrans(
                pya.Trans.R0, min(bus_x) / dbu, (min(bus_y) + max(bus_y)) / 2 / dbu
            ),
        )
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = pya.DTrans(
            pya.Trans.R90, max(bus_x) / dbu, (min(bus_y) + max(bus_y)) / 2 / dbu
        )
        pin = pya.Path(
            [pya.DPoint(0, pin_length / 2), pya.DPoint(0, -pin_length / 2)], bus_w / dbu
        )
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text(
            "opt2",
            pya.DTrans(
                pya.Trans.R90, max(bus_x) / dbu, (min(bus_y) + max(bus_y)) / 2 / dbu
            ),
        )
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Dev box
        # Encapsulate the pcell within a box for error checking
        w = self.w
        dev = pya.DBox(-(r + w), r + w + cg + bus_w + 0.5, r + w, -(r + w) - 0.5)
        shapes(LayerDevRecN).insert(dev)



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from . import *


class MMI_2x2(pya.PCellDeclarationHelper):
    """ """

    def __init__(self):
        # Important: initialize the super class
        super(MMI_2x2, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("w_mmi", self.TypeDouble, "Waveguide Width", default=6.0)
        self.param("w_a", self.TypeDouble, "Access waveguide Width", default=1.5)
        self.param("L_mmi", self.TypeDouble, "MMI Length", default=130.5)
        self.param("L_a", self.TypeDouble, "Access waveguide Length", default=10.0)
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "MMI_2x2"

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # test by Lukas, to see if it breaks the PCell
        # print(1/0)

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.silayer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TextLayerN = ly.layer(self.textl)

        w = int(round(self.w / dbu))
        w_mmi = int(round(self.w_mmi / dbu))
        w_a = int(round(self.w_a / dbu))
        L_mmi = int(round(self.L_mmi / dbu))
        L_a = int(round(self.L_a / dbu))

        # Create the MMI
        mmi = Box(-L_mmi / 2, -w_mmi / 2, L_mmi / 2, w_mmi / 2)
        shapes(LayerSiN).insert(mmi)

        # Draw taper access waveguides
        # Create a sub-cell
        cell = self.cell.layout().create_cell("MMI_access_waveguide")
        pts = []
        pts.append(Point.from_dpoint(DPoint(0, w_a / 2)))
        pts.append(Point.from_dpoint(DPoint(L_a, w / 2)))
        pts.append(Point.from_dpoint(DPoint(L_a, -w / 2)))
        pts.append(Point.from_dpoint(DPoint(0, -w_a / 2)))
        polygon = Polygon(pts)
        cell.shapes(LayerSiN).insert(polygon)
        # place the cell in the top cell
        self.cell.insert(
            CellInstArray(cell.cell_index(), Trans(Trans.R0, L_mmi / 2, w_a))
        )
        self.cell.insert(
            CellInstArray(cell.cell_index(), Trans(Trans.R0, L_mmi / 2, -w_a))
        )
        self.cell.insert(
            CellInstArray(cell.cell_index(), Trans(Trans.R180, -L_mmi / 2, w_a))
        )
        self.cell.insert(
            CellInstArray(cell.cell_index(), Trans(Trans.R180, -L_mmi / 2, -w_a))
        )

        pin(
            500,
            "pin1",
            Trans(Trans.R180, -L_mmi / 2 - L_a, w_a),
            LayerPinRecN,
            dbu,
            self.cell,
        )
        pin(
            500,
            "pin2",
            Trans(Trans.R180, -L_mmi / 2 - L_a, -w_a),
            LayerPinRecN,
            dbu,
            self.cell,
        )
        pin(
            500,
            "pin3",
            Trans(Trans.R0, L_mmi / 2 + L_a, w_a),
            LayerPinRecN,
            dbu,
            self.cell,
        )
        pin(
            500,
            "pin4",
            Trans(Trans.R0, L_mmi / 2 + L_a, -w_a),
            LayerPinRecN,
            dbu,
            self.cell,
        )

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        dev = Box(-L_mmi / 2 - L_a, w_mmi / 2 + w, L_mmi / 2 + L_a, -w_mmi / 2 - w)
        shapes(LayerDevRecN).insert(dev)

        # print("Done drawing the layout for - mmi" )



from pya import *
import pya

def linspace_without_numpy(low, high, length):
    step = ((high-low) * 1.0 / length)
    return [low+i*step for i in range(length)]

class ebeam_GC_SiN_TE_1310_8deg(pya.PCellDeclarationHelper):
  """
  This design is based on the Universal Grating Coupler PCell implementation. Which details can be found below:
    Universal Grating Coupler PCell implementation.
    Analytical design based on "Grating Coupler Design Based on Silicon-On-Insulator", Yun Wang (2013). Master's Thesis, University of British Columbia, Canada
    Some PCell implementation adapted from the SiEPIC_EBeam library by Dr. Lukas Chrostowski, University of British Columbia, Canada
    Orignal script written by Timothy Richards (Simon Fraser University, BC, Canada) and Adam DeAbreu (Simon Fraser University, BC, Canada)
    Modified to use UGC method presented in : Yun Wang, Jonas Flueckiger, Charlie Lin, Lukas Chrostowski, "Universal grating coupler design," Proc. SPIE 8915, Photonics North 2013, 89150Y (11 October 2013), 
    Transfered from Mentor Graphics to KLayout by Connor Mosquera (University of British Columbia, Canada)

  This design was adapted by: Hang (Bobby) Zou from Si to SiN platform, with default parameters set as best device parameters with real fabrication run
  Date Edited: 2024-08-28
  
  """

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_GC_SiN_TE_1310_8deg, self).__init__()
    from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.TECHNOLOGY = TECHNOLOGY
    
    # declare the parameters  
    self.param("wavelength", self.TypeDouble, "Design Wavelength (micron)", default = 1.31)  
    self.param("Si_thickness", self.TypeDouble, "SiN Thickness (micron)", default = 0.35)
    self.param("etch_depth", self.TypeDouble, "Etch Depth (micron)", default = 0.4)
    self.param("pol", self.TypeString, "Polarization", default = "TE")
    self.param("n_t", self.TypeDouble, "Cladding Index", default = 1.4718)
    self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default = 36.0)
    self.param("grating_length", self.TypeDouble, "Grating Length (micron)", default = 30.0)
    self.param("taper_length", self.TypeDouble, "Taper Length (micron)", default = 20)
    self.param("dc", self.TypeDouble, "Duty Cylce", default = 0.55)
    self.param('pitch', self.TypeDouble, "Pitch (micron)", default = 0.94)
    self.param("t", self.TypeDouble, "Waveguide Width (micron)", default = 0.75)
    self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default = 8.121)
    
    # Layer parameters
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['SiN'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "EBeam_GC_SiN_1310_8deg_%.1f-%.2f-%.2f-%.2f-%.2f-%.2f" % \
    (self.wavelength, self.theta_c, self.dc, self.angle_e, self.taper_length, self.t)
    
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    TECHNOLOGY = self.TECHNOLOGY

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerSiSPN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    LayerTextN = TECHNOLOGY['Text']
    
    ######## effective index function ##########
    
    def effective_index(wl = self.wavelength, etch_depth = self.etch_depth, Si_thickness = self.Si_thickness, n_t = self.n_t, pol = self.pol, dc = self.dc):
      
      from math import pi, cos, sin, log, sqrt, tan
      from SiEPIC.utils import points_per_circle
      
      point =1001
      n_0 = n_t # Top cladding
      n_1 = 0
      n_3 = 1.4718 # Bottom Cladding
      # if 0:  # Silicon
      #     n_2 = sqrt(7.9874+(3.68*pow(3.9328,2)*pow(10,30))/((pow(3.9328,2)*pow(10,30)-pow(2*3.14*3*pow(10,8)/(wl*pow(10,-6)),2))))  # Silicon wavelength-dependant index of refraction
      # if 1:  # Silicon Nitride
      n_2 = 2.001667  # approximate  
      delta = n_0 - n_3 
      t = Si_thickness
      t_slot = t - etch_depth
      k_0 = 2*pi/wl
      
      b_0 = linspace_without_numpy(0, 0, point-1)
      te_0 = linspace_without_numpy(0, 0, point-1)
      te_1 = linspace_without_numpy(0, 0, point-1)
      tm_0 = linspace_without_numpy(0, 0, point-1)
      tm_1 = linspace_without_numpy(0, 0, point-1)
      h_0 = linspace_without_numpy(0, 0, point-1)
      q_0 = linspace_without_numpy(0, 0, point-1)
      p_0 = linspace_without_numpy(0, 0, point-1)
      qbar_0 = linspace_without_numpy(0, 0, point-1)
      pbar_0 = linspace_without_numpy(0, 0, point-1)
      
      # calculating neff for the silicon layer
      if delta<0:
        n_1 = n_3
      else:
        n_1 = n_0
        
      for ii in range(0,point-1):
        b_0[ii] = n_1*k_0 + (n_2 - n_1)*k_0/(point-1)*ii   # copied from .ample UGC: should this be point-1?
        
        h_0[ii] = sqrt( abs(pow(n_2*k_0,2) - pow(b_0[ii],2)));	
        q_0[ii] = sqrt( abs(pow(b_0[ii],2) - pow(n_0*k_0,2)));
        p_0[ii] = sqrt( abs(pow(b_0[ii],2) - pow(n_3*k_0,2)));
        
        pbar_0[ii] = pow(n_2/n_3, 2)*p_0[ii]
        qbar_0[ii] = pow(n_2/n_0, 2)*q_0[ii]
        
      # calculating neff for TE mode
      if pol == "TE":
        for ii in range(0, point-1):
          te_0[ii] =  tan( h_0[ii]*t )-(p_0[ii]+q_0[ii])/h_0[ii]/(1-p_0[ii]*q_0[ii]/pow(h_0[ii],2)) 
          te_1[ii] =  tan( h_0[ii]*t_slot )-(p_0[ii]+q_0[ii])/h_0[ii]/(1-p_0[ii]*q_0[ii]/pow(h_0[ii],2))
        
        abs_te_0 = [abs(x) for x in te_0] 
        abs_te_1 = [abs(x) for x in te_1] 
        index_TE_0 = abs_te_0.index(min(abs_te_0))
        index_TE_1 = abs_te_1.index(min(abs_te_1))
        nTE_0 = b_0[index_TE_0]/k_0
        nTE_1 = b_0[index_TE_1]/k_0
        nTE_1 = n_3
        
        while (nTE_0<1.5 or nTE_0>3):
          abs_te_0[index_TE_0] = 100
          index_TE_0 = abs_te_0.index(min(abs_te_0))
          nTE_0 = b_0[index_TE_0]/k_0
          
        '''  
        while (nTE_1<1.5 or nTE_1>3):
          abs_te_1[index_TE_1] = 100
          index_TE_1 = abs_te_1.index(min(abs_te_1))
          nTE_1 = b_0[index_TE_1]/k_0
        '''
        
        ne=dc*nTE_0+(1-dc)*nTE_1
        
      # calculating neff for TE mode
      elif pol == "TM":
        for ii in range(0, point-1):
          tm_0[ii] =  tan( h_0[ii]*t )-(pbar_0[ii]+qbar_0[ii])/h_0[ii]/(1-pbar_0[ii]*qbar_0[ii]/pow(h_0[ii],2)) 
          tm_1[ii] =  tan( h_0[ii]*t_slot )-(pbar_0[ii]+qbar_0[ii])/h_0[ii]/(1-pbar_0[ii]*qbar_0[ii]/pow(h_0[ii],2))
        
        abs_tm_0 = [abs(x) for x in tm_0] 
        abs_tm_1 = [abs(x) for x in tm_1] 
        index_TM_0 = abs_tm_0.index(min(abs_tm_0))
        index_TM_1 = abs_tm_1.index(min(abs_tm_1))
        nTM_0 = b_0[index_TM_0]/k_0
        nTM_1 = b_0[index_TM_1]/k_0
        
        while (nTM_0<1.5 or nTM_0>3):
          abs_tm_0[index_TM_0] = 100
          index_TM_0 = abs_tm_0.index(min(abs_tm_0))
          nTM_0 = b_0[index_TM_0]/k_0
          
        while (nTM_1<1.5 or nTM_1>3):
          abs_tm_1[index_TM_1] = 100
          index_TM_1 = abs_tm_1.index(min(abs_tm_1))
          nTM_1 = b_0[index_TM_1]/k_0
        
        ne=dc*nTM_0+(1-dc)*nTM_1
      
      else:
        print('Please type TE or TM for polarization')
      
      return ne
    #####################################
    
    from math import pi, cos, sin, log, sqrt, tan
    from SiEPIC.utils import points_per_circle
    
    lambda_0 = self.wavelength                   ##um wavelength of light
    
    n_e = effective_index()
    #n_e = 1.6975992
    ne_fiber = 1 # effective index of the mode in the air
    period = self.wavelength/(n_e - sin(pi/180*self.theta_c)*ne_fiber)
    period = self.pitch


    # Geometry
    wh = period*self.dc                   ##thick grating
    
    gc_number = int(round(self.grating_length/period)) ##number of periods
    e = self.n_t*sin((pi/180)*self.theta_c)/n_e
    N = round(self.taper_length*(1+e)*n_e/lambda_0) ##allows room for the taper

    start = (pi - (pi/180)*self.angle_e/2)
    stop = (pi + (pi/180)*self.angle_e/2)

    radius_1 = 999 # random number so we know there is an error if radius_1 is not overwritten
    radius_2 = 999
    # Draw coupler grating.
    for j in range(gc_number):

        # number of points in the arcs:
        # calculate such that the vertex & edge placement error is < 0.5 nm.
        #   see "SiEPIC_EBeam_functions - points_per_circle" for more details
        radius = N*lambda_0 / (n_e*( 1 - e )) + j*period
        
        if j == 0:
            radius_1 = radius

        if j == 1:
            radius_2 = radius

        seg_points = int(points_per_circle(radius/dbu, dbu=dbu)/360.*self.angle_e) # number of points grating arc
        theta_up = []
        for m in range(seg_points+1):    
          theta_up = theta_up + [start + m*(stop-start)/seg_points]
        theta_down = theta_up[::-1]
      
        ##big one
        r_up = []
        r_down = []
        for k in range(len(theta_up)):
          r_up = r_up + [N*lambda_0 / (n_e*( 1 - e*cos(float(theta_up[k])) )) + j*period+period*(1-self.dc)]
        r_down = r_up[::-1]
    
        xr = []
        yr = []
        for k in range(len(theta_up)):
          xr = xr + [r_up[k]*cos(theta_up[k])]
          yr = yr + [r_up[k]*sin(theta_up[k])]
    
        xl = []
        yl = []
        for k in range(len(theta_down)):
          xl = xl + [(r_down[k] + wh)*cos(theta_down[k])]
          yl = yl + [(r_down[k] + wh)*sin(theta_down[k])]
    
        x = xr + xl
        y = yr + yl
        
        pts = []
        for i in range(len(x)):
            pts.append(Point.from_dpoint(DPoint(x[i]/dbu, y[i]/dbu)))
        
        polygon = Polygon(pts)
        shapes(LayerSiN).insert(polygon)
      
    # Taper section
    r_up = []
    r_down = []
    for k in range(len(theta_up)):  
      r_up = r_up + [N*lambda_0 / (n_e*( 1 - e*cos(float(theta_up[k])) ))]
    r_down = r_up[::-1]
     
    xl = []
    yl = []  
    for k in range(len(theta_down)):
      xl = xl + [(r_down[k])*cos(theta_down[k])]
      yl = yl + [(r_down[k])*sin(theta_down[k])]
      
    yr = [self.t/2., -self.t/2.]
    
    yl_abs = []
    for k in range(len(yl)):
      yl_abs = yl_abs + [abs(yl[k])]
    
    y_max = max(yl_abs)
    iy_max = yl_abs.index(y_max)
      
    L_o = (y_max - self.t/2)/tan((pi/180)*self.angle_e/2)
      
    xr = [0, 0]
     
    x = xr + xl
    y = yr + yl  

    pts = []
    for i in range(len(x)):
      pts.append(Point.from_dpoint(DPoint(x[i]/dbu, y[i]/dbu)))
    
    polygon = Polygon(pts)
    shapes(LayerSiN).insert(polygon)


    # Pin on the waveguide:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    x = 0
    t = Trans(Trans.R0, x,0)
    pin = Path([Point(-pin_length/2,0), Point(pin_length/2,0)], self.t/dbu)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("opt1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Reference information
    t = Trans(Trans.R0, 0,-4000)
    text = Text ("Ref: 'Universal grating coupler design'\nhttps://doi.org/10.1117/12.2042185\nPCell implementation by: Yun Wang, Timothy Richards, Adam DeAbreu,\nJonas Flueckiger, Charlie Lin, Lukas Chrostowski, Connor Mosquera\n Adapted by Hang Bobby Zou", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.5/dbu
    shape.text_halign=2 # right alignment

    t = Trans(Trans.R0, 0,4000)
    text = Text ("Wavelength: %s\nIncident Angle: %s\nPolarization: %s\nSiN thickness: %s\nSiN etch depth: %s\nDuty Cycle: %s\nPitch: %s\nRadius_1: %s\nRadius_2: %s\nn_eff: %s\ne: %s\nN: %s\n" % (self.wavelength, self.theta_c, self.pol, self.Si_thickness, self.etch_depth, self.dc, self.pitch, radius_1, radius_2,n_e,e,N), t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.5/dbu
    shape.text_halign=2 # right alignment
    shape.text_valign=2 # bottom alignment

    # Device recognition layer 
    yr = sin(start) * (N*lambda_0 / ( n_e*( 1 - e*cos(float(start)) )) + gc_number*period )
    box1 = Box(-(self.grating_length+self.taper_length)/dbu-pin_length*2, yr/dbu, 0, -yr/dbu)
    shapes(LayerDevRecN).insert(box1)       



import pya
from . import *
from pya import *


class Waveguide_bump(pya.PCellDeclarationHelper):
    """
    Create a waveguide with a "bump" consisting of four bends,
    to implement an extra length as compared to a straight waveguide.
    Input:
     - delta_length
     - waveguide_type from Waveguides.XML
    Calculates the theta for the four arcs
    Uses waveguide parameters
    Increases the bend radius in case of Bezier curves
    """

    def __init__(self):
        # Important: initialize the super class
        super(Waveguide_bump, self).__init__()
        from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech

        TECHNOLOGY = get_technology_by_name("EBeam")
        self.technology_name = "EBeam"

        # Load all waveguides types
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)

        # find SiN waveguide
        names = [w["name"] for w in self.waveguide_types if "SiN" in w["name"]]
        if not names:
            raise Exception("Waveguides.XML does not have a SiN waveguide.")

        # declare the parameters
        p = self.param(
            "waveguide_type", self.TypeList, "Waveguide Type", default=names[0]
        )
        for wa in self.waveguide_types:
            p.add_choice(wa["name"], wa["name"])

        # declare the parameters
        self.param(
            "delta_length", self.TypeDouble, "Waveguide length difference", default=0.2
        )
        self.param("max_theta", self.TypeDouble, "Maximum angle (degrees)", default=149)
        self.param(
            "radius_mult",
            self.TypeDouble,
            "Waveguide radius multiplier for Bezier bends",
            default=2,
        )
        self.param(
            "radius",
            self.TypeDouble,
            "Waveguide Radius (microns)",
            default=self.waveguide_types[0]["radius"],
            readonly=True,
        )
        self.param(
            "wg_width",
            self.TypeDouble,
            "Waveguide width (microns)",
            default=0.5,
            readonly=True,
        )
        self.param(
            "theta",
            self.TypeDouble,
            "Waveguide angle (degrees)",
            default=5,
            hidden=True,
        )
        self.param(
            "length",
            self.TypeDouble,
            "Regular Waveguide length",
            default=10.0,
            hidden=True,
        )

        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("text", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Waveguide_bump_%s-%.3f" % (self.length, self.wg_width)

    def coerce_parameters_impl(self):
        params = [t for t in self.waveguide_types if t["name"] == self.waveguide_type]
        if not params:
            raise Exception("Waveguides.XML not correctly configured")

        if "bezier" in params[0]:
            # chose a radius that is larger to avoid loss
            mult = self.radius_mult
        else:
            mult = 1
        self.radius = float(params[0]["radius"]) * mult
        self.wg_width = float(params[0]["width"])

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")
        params = [t for t in self.waveguide_types if t["name"] == self.waveguide_type]
        layer = [wg["layer"] for wg in params[0]["component"]]
        try:
            layer.remove("DevRec")
        except:
            pass
        if not layer:
            raise Exception("Waveguides.XML not correctly configured")
        LayerSiN = ly.layer(TECHNOLOGY[layer[0]])
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.text)

        from math import pi, cos, sin
        from SiEPIC.utils import arc_wg

        x = 0
        y = 0

        from SiEPIC.extend import to_itype

        # Waveguide parameters:
        w = to_itype(self.wg_width, self.layout.dbu)
        r = to_itype(self.radius, dbu)

        # Solve for the value of theta that gives us the desired delta_length
        # four arcs with the same radius and angle
        # delta_length = waveguide_length - device length
        # delta_length = 2*pi*r*(4*theta/360.0) - 4*r*sin(theta/180.0*pi)
        """
    Code from ChatGPT:
    Lukas: I want to solve the following equation for theta, 
            given delta_length and r, using Python: 
            delta_length = 2*pi*r*(4*theta/360.0) - 4*r*sin(theta/180.0*pi)
    ChatGPT:
    To solve the equation for theta using Python, 
    you can use a numerical optimization method like the bisection method, 
    Newton's method, or the secant method. 
    Here's an example implementation using the bisection method:
    Here, the f function represents the equation we want to solve, 
    and the solve_theta function uses the bisection method to find 
    the root of this equation within a given tolerance. 
    The a and b variables represent the endpoints of the initial interval, 
    and the tol variable represents the desired tolerance for the solution.
    To use this function, you can call solve_theta with your values 
    for delta_length and r.
    This will output the value of theta that satisfies the equation. 
    Note that the solution will be in degrees, since that's the unit 
    used in the equation. If you need the solution in radians, 
    you can convert it using theta_rad = theta / 180 * math.pi.
    """

        def f(theta, delta_length, r):
            import math

            return (
                delta_length
                - 2 * math.pi * r * (4 * theta / 360.0)
                + 4 * r * math.sin(theta / 180.0 * math.pi)
            )

        def solve_theta(delta_length, r):
            a, b = 0, 180
            tol = 1e-6
            while abs(b - a) > tol:
                c = (a + b) / 2
                if f(a, delta_length, r) * f(c, delta_length, r) < 0:
                    b = c
                else:
                    a = c
            return (a + b) / 2

        theta = solve_theta(self.delta_length, r * dbu)

        # Check for angle
        if theta > self.max_theta:
            theta = self.max_theta
            self.delta_length = 2 * pi * r * dbu * (
                4 * theta / 360.0
            ) - 4 * r * dbu * sin(theta / 180.0 * pi)
            self.coerce_parameters_impl()

        # length of the waveguide and device:
        waveguide_length = 2 * pi * r * (4 * theta / 360.0)
        length = 4 * r * sin(theta / 180.0 * pi)
        delta_length = waveguide_length - length

        t = Trans(Trans.R0, x, round(y + r))
        self.cell.shapes(LayerSiN).insert(
            arc_wg(r, w, 270.0, 270.0 + theta, dbu=dbu).transformed(t)
        )
        t = Trans(
            Trans.R0,
            round(x + length / 2),
            round(y - r + 2 * r * (1 - cos(theta / 180.0 * pi))),
        )
        self.cell.shapes(LayerSiN).insert(
            arc_wg(r, w, 90.0 - theta, 90.0 + theta, dbu=dbu).transformed(t)
        )
        t = Trans(Trans.R0, round(x + length), round(y + r))
        self.cell.shapes(LayerSiN).insert(
            arc_wg(r, w, 270.0 - theta, 270, dbu=dbu).transformed(t)
        )

        # Create the pins on the waveguides, as short paths:
        from SiEPIC.utils.layout import make_pin

        make_pin(self.cell, "pin1", [x, y], w, LayerPinRecN, 180)
        make_pin(self.cell, "pin2", [x + length * dbu, y], w, LayerPinRecN, 0)

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        cml = params[0]["CML"]
        if cml:
            text = Text("Lumerical_INTERCONNECT_library=Design kits/%s" % cml, t)
            shape = shapes(LayerDevRecN).insert(text)
            shape.text_size = 0.1 / dbu
            t = Trans(Trans.R0, 0, w * 2)
        model = params[0]["model"]
        if model:
            text = Text("Component=%s" % model, t)
            shape = shapes(LayerDevRecN).insert(text)
            shape.text_size = 0.1 / dbu
            t = Trans(Trans.R0, 0, -w * 2)
        text = Text(
            "Spice_param:wg_length=%.3fu wg_width=%.3fu"
            % (waveguide_length * dbu, self.wg_width),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        t = Trans(Trans.R0, self.length / 6, -w * 2)
        text = Text("dL = %.4f um" % (delta_length * dbu), t)
        shape = shapes(LayerTextN).insert(text)
        #    shape.text_size = w*1.2
        shape.text_size = length * 0.06

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        box1 = Box(0, -w * 3, length, w * 3 + (2 * r * (1 - cos(theta / 180.0 * pi))))
        shapes(LayerDevRecN).insert(box1)

        # print("SiEPIC EBeam: Waveguide_bump complete.")



import os
import sys
import SiEPIC
from SiEPIC.utils import get_technology, get_technology_by_name
from pya import *
import math

path = os.path.dirname(os.path.abspath(__file__))


def linspace_without_numpy(low, high, length):
    step = (high - low) * 1.0 / length
    return [low + i * step for i in range(length)]


def pin(w, pin_text, trans, LayerPinRecN, dbu, cell):
    """
    w: Waveguide Width, e.g., 500 (in dbu)
    pin_text: Pin Text, e.g., "pin1"
    trans: e.g., trans =  Trans(0, False, 0, 0)  - first number is 0, 1, 2, or 3.
    pinrec: PinRec Layer, e.g., layout.layer(TECHNOLOGY['PinRec']))
    devrec: DevRec Layer, e.g., layout.layer(TECHNOLOGY['DevRec']))
    """

    # Create the pin, as short paths:
    from SiEPIC._globals import PIN_LENGTH

    pin = trans * Path([Point(-PIN_LENGTH / 2, 0), Point(PIN_LENGTH / 2, 0)], w)
    cell.shapes(LayerPinRecN).insert(pin)
    text = Text(pin_text, trans)
    shape = cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = w * 0.8

    # print("Done drawing the layout for - pin" )



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class taper_bezier(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the strip waveguide taper using Bezier curve.
    """

    def __init__(self):
        # Important: initialize the super class
        super(taper_bezier, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # Button to launch separate window
        self.param(
            "documentation", self.TypeCallback, "Open documentation in web browser"
        )
        #    self.param("simulation", self.TypeCallback, "Launch simulation GUI")

        # declare the parameters
        self.param(
            "wg_width1", self.TypeDouble, "Waveguide Width1 (micron)", default=0.5
        )
        self.param("wg_width2", self.TypeDouble, "Waveguide Width2 (micron)", default=3)
        self.param(
            "wg_length", self.TypeDouble, "Waveguide Length (micron)", default=10
        )
        self.param("a", self.TypeDouble, "Bezier parameter, a (0-1)", default=0.37)
        self.param("b", self.TypeDouble, "Bezier parameter, b (0-1)", default=0.37)
        self.param(
            "accuracy", self.TypeDouble, "Curve accuracy (micron)", default=0.005
        )
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        # hidden parameters, can be used to query this component:
        self.param(
            "p1",
            self.TypeShape,
            "DPoint location of pin1",
            default=Point(-10000, 0),
            hidden=True,
            readonly=True,
        )
        self.param(
            "p2",
            self.TypeShape,
            "DPoint location of pin2",
            default=Point(0, 10000),
            hidden=True,
            readonly=True,
        )

    def callback(self, layout, name, states):
        """Callback for PCell, to launch documentation viewer
        https://www.klayout.de/doc/code/class_PCellDeclaration.html#method9
        """
        if name == "documentation":
            url = "https://github.com/SiEPIC/SiEPIC_EBeam_PDK/blob/master/Documentation/Taper/Summary_Taper.pdf"
            import webbrowser

            webbrowser.open_new(url)

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "taper_bezier("
            + (
                "%.3f-%.3f-%.3f-%.3f-%.3f"
                % (self.wg_width1, self.wg_width2, self.wg_length, self.a, self.b)
            )
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce(self, layout, layers, parameters, cell):
        from packaging import version
        import SiEPIC

        if version.parse(SiEPIC.__version__) < version.parse("0.5.11"):
            raise Exception(
                "Errors",
                "This PCell requires SiEPIC-Tools version 0.5.11 or greater. You have %s"
                % SiEPIC.__version__,
            )

        from SiEPIC.utils.geometry import bezier_cubic
        from SiEPIC.extend import to_itype
        from SiEPIC.utils.layout import make_pin

        """
    coerce parameters (make consistent)
    """
        self._layers = layers
        self.cell = cell
        self._param_values = parameters
        self.layout = layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        w1 = to_itype(self.wg_width1, dbu)
        w2 = to_itype(self.wg_width2, dbu)
        length = to_itype(self.wg_length, dbu)

        wg_Dpts = (
            bezier_cubic(
                pya.DPoint(0, self.wg_width1 / 2),
                pya.DPoint(self.wg_length, self.wg_width2 / 2),
                0,
                0,
                self.a,
                self.b,
                accuracy=self.accuracy,
            )
            + bezier_cubic(
                pya.DPoint(0, -self.wg_width1 / 2),
                pya.DPoint(self.wg_length, -self.wg_width2 / 2),
                0,
                0,
                self.a,
                self.b,
                accuracy=self.accuracy,
            )[::-1]
        )
        wg_polygon = pya.DPolygon(wg_Dpts)
        cell.shapes(LayerSi).insert(wg_polygon)

        # Create the pins on the waveguides, as short paths:
        make_pin(self.cell, "opt1", [0, 0], w1, LayerPinRecN, 180)
        make_pin(self.cell, "opt2", [length, 0], w2, LayerPinRecN, 0)

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        path = pya.Path([pya.Point(0, 0), pya.Point(length, 0)], w2 + w1 * 2)
        cell.shapes(LayerDevRecN).insert(path.simple_polygon())

        # Compact model information
        t = pya.Trans(pya.Trans.R0, w1 / 10, 0)
        text = Text("Lumerical_INTERCONNECT_library=None", t)
        shape = cell.shapes(LayerDevRecN).insert(text)
        shape.text_size = length / 100
        t = pya.Trans(pya.Trans.R0, length / 10, w1 / 4)
        text = pya.Text("Component=taper_bezier", t)
        shape = cell.shapes(LayerDevRecN).insert(text)
        shape.text_size = length / 100
        t = pya.Trans(pya.Trans.R0, length / 10, w1 / 2)
        text = pya.Text(
            "Spice_param:wg_width1=%.3fu wg_width2=%.3fu wg_length=%.3fu"
            % (self.wg_width1, self.wg_width2, self.wg_length),
            t,
        )
        shape = cell.shapes(LayerDevRecN).insert(text)
        shape.text_size = length / 100

        return (
            "taper_bezier("
            + ("%.3f-%.3f-%.3f" % (self.wg_width1, self.wg_width2, self.wg_length))
            + ")"
        )


class test_lib(Library):
    def __init__(self):
        tech = "EBeam"
        library = tech + "test_lib"
        self.technology = tech
        self.layout().register_pcell("taper_bezier", taper_bezier())
        self.register(library)


if __name__ == "__main__":
    print("Test layout for: Taper Bezier")

    import siepic_ebeam_pdk
    from SiEPIC.utils.layout import new_layout

    # load the test library, and technology
    t = test_lib()
    tech = t.technology

    # Create a new layout for the chip floor plan
    topcell, ly = new_layout(tech, "test", GUI=True, overwrite=True)

    # instantiate the cell
    library = tech + "test_lib"

    # Create for several combinations
    params = [
        {"wg_width1": 0.5, "wg_width2": 1, "wg_length": 10, "a": 0.37, "b": 0.37},
        {"wg_width1": 0.5, "wg_width2": 2, "wg_length": 20, "a": 0.37, "b": 0.37},
        {"wg_width1": 0.5, "wg_width2": 3, "wg_length": 10, "a": 0.37, "b": 0.37},
        {"wg_width1": 0.5, "wg_width2": 3, "wg_length": 20, "a": 0.37, "b": 0.37},
        {"wg_width1": 0.5, "wg_width2": 3, "wg_length": 20, "a": 0.2, "b": 0.2},
        {"wg_width1": 0.5, "wg_width2": 3, "wg_length": 20, "a": 0.7, "b": 0.7},
        {"wg_width1": 0.5, "wg_width2": 3, "wg_length": 20, "a": 0.7, "b": 0.3},
        {"wg_width1": 0.5, "wg_width2": 3, "wg_length": 40, "a": 0.37, "b": 0.37},
    ]
    xmax = 0
    y = 0
    x = xmax
    for p in params:
        print(p)
        pcell = ly.create_cell("taper_bezier", library, p)
        t = pya.Trans(pya.Trans.R0, x, y - pcell.bbox().bottom)
        inst = topcell.insert(pya.CellInstArray(pcell.cell_index(), t))
        y += pcell.bbox().height() + 2000
        xmax = max(xmax, x + inst.bbox().width())

    # Save
    import os
    from SiEPIC.scripts import export_layout

    path = os.path.dirname(os.path.realpath(__file__))
    filename = os.path.splitext(os.path.basename(__file__))[0]
    file_out = export_layout(topcell, path, filename, format="oas", screenshot=True)

    # Display in KLayout
    from SiEPIC._globals import Python_Env

    if Python_Env == "Script":
        from SiEPIC.utils import klive

        klive.show(file_out, technology=tech, keep_position=True)



import pya
from pya import *


class ebeam_dc_halfring_straight(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the ebeam_dc_halfring_straight.
    Consists of a half-ring with 1 waveguides.
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_dc_halfring_straight, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("r", self.TypeDouble, "Radius", default=10)
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("g", self.TypeDouble, "Gap", default=0.2)
        self.param("Lc", self.TypeDouble, "Coupler Length", default=0.0)
        self.param(
            "orthogonal_identifier",
            self.TypeInt,
            "Orthogonal identifier (1=TE, 2=TM)",
            default=1,
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "ebeam_dc_halfring_straight(R="
            + ("%.3f" % self.r)
            + ",g="
            + ("%g" % (1000 * self.g))
            + ",Lc="
            + ("%g" % (1000 * self.Lc))
            + ",orthogonal_identifier="
            + ("%s" % self.orthogonal_identifier)
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        from SiEPIC.utils import arc_wg_xy
        from SiEPIC._globals import PIN_LENGTH

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSiN = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TextLayerN = ly.layer(self.textl)

        w = int(round(self.w / dbu))
        r = int(round(self.r / dbu))
        g = int(round(self.g / dbu))
        Lc = int(round(self.Lc / dbu))

        # draw the half-circle
        x = 0
        y = r + w + g
        from SiEPIC import __version__

        if __version__ > "0.5.0":
            self.cell.shapes(LayerSiN).insert(
                arc_wg_xy(x - Lc / 2, y, r, w, 180, 270, dbu=dbu)
            )  # dbu argument introduced in SiEPIC 0.5.1
            self.cell.shapes(LayerSiN).insert(
                arc_wg_xy(x + Lc / 2, y, r, w, 270, 360, dbu=dbu)
            )
        else:
            self.cell.shapes(LayerSiN).insert(arc_wg_xy(x - Lc / 2, y, r, w, 180, 270))
            self.cell.shapes(LayerSiN).insert(arc_wg_xy(x + Lc / 2, y, r, w, 270, 360))

        # Pins on the top side:
        pin = Path(
            [
                Point(-r - Lc / 2, y - PIN_LENGTH / 2),
                Point(-r - Lc / 2, y + PIN_LENGTH / 2),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, -r - Lc / 2, y)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(r + Lc / 2, y - PIN_LENGTH / 2),
                Point(r + Lc / 2, y + PIN_LENGTH / 2),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, r + Lc / 2, y)
        text = Text("pin4", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        if Lc > 0:
            wg1 = Box(-Lc / 2, -w / 2 + w + g, Lc / 2, w / 2 + w + g)
            shapes(LayerSiN).insert(wg1)

        # Create the waveguide
        wg1 = Box(-r - w / 2 - w - Lc / 2, -w / 2, r + w / 2 + w + Lc / 2, w / 2)
        shapes(LayerSiN).insert(wg1)

        # Pins on the bus waveguide side:
        pin = Path(
            [
                Point(-r - w / 2 - w + PIN_LENGTH / 2 - Lc / 2, 0),
                Point(-r - w / 2 - w - PIN_LENGTH / 2 - Lc / 2, 0),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, -r - w / 2 - w - Lc / 2, 0)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(r + w / 2 + w - PIN_LENGTH / 2 + Lc / 2, 0),
                Point(r + w / 2 + w + PIN_LENGTH / 2 + Lc / 2, 0),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, r + w / 2 + w + Lc / 2, 0)
        text = Text("pin3", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Merge all the waveguide shapes, to avoid any small gaps
        layer_temp = self.layout.layer(LayerInfo(913, 0))
        shapes_temp = self.cell.shapes(layer_temp)
        ShapeProcessor().merge(
            self.layout, self.cell, LayerSiN, shapes_temp, True, 0, True, True
        )
        self.cell.shapes(LayerSiN).clear()
        shapes_SiN = self.cell.shapes(LayerSiN)
        ShapeProcessor().merge(
            self.layout, self.cell, layer_temp, shapes_SiN, True, 0, True, True
        )
        self.cell.shapes(layer_temp).clear()

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        dev = Box(-r - w / 2 - w - Lc / 2, -w / 2 - w, r + w / 2 + w + Lc / 2, y)
        shapes(LayerDevRecN).insert(dev)

        # Compact model information
        t = Trans(Trans.R0, -r, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = self.r * 0.03 / dbu

        t = Trans(Trans.R0, -r, r / 4)
        text = Text("Component=ebeam_dc_halfring_straight", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = self.r * 0.03 / dbu

        t = Trans(Trans.R0, -r, r / 2)
        text = Text(
            "Spice_param:wg_width=%.3g gap=%.3g radius=%.3g Lc=%.3g orthogonal_identifier=%s"
            % (
                self.w * 1e-6,
                self.g * 1e-6,
                self.r * 1e-6,
                self.Lc * 1e-6,
                self.orthogonal_identifier,
            ),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = self.r * 0.03 / dbu

        # print("Done drawing the layout for - ebeam_dc_halfring_straight: %.3f-%g" % ( self.r, self.g) )
